/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <windows.h>
#include <shlobj.h>
#include "resource.h"
#include "main.h"
#include "translation.h"
#include "Gui_SettingsSheets.h"
#include "config.h"
#include "nSX2.h"

////////////////////////////////////////////////////////////////////////////////
/*  Log Messages Funtions  */

/* Saves Log vars in nSX2 Config File*/
void saveLogVars()
{
    char TempMessage[100] ; 
    
    sprintf(TempMessage,"%d",debug_gs);
    WriteCfgString("Logging","GS",TempMessage);
    
    sprintf(TempMessage,"%d",debug_dma);
    WriteCfgString("Logging","DMA",TempMessage);
    
    sprintf(TempMessage,"%d",debug_systemcall);
    WriteCfgString("Logging","SYSCALL",TempMessage);
    
    sprintf(TempMessage,"%d",debug_gif);
    WriteCfgString("Logging","GIF",TempMessage);
    
    sprintf(TempMessage,"%d",debug_intc);
    WriteCfgString("Logging","INTC",TempMessage);
    
    sprintf(TempMessage,"%d",debug_timer);
    WriteCfgString("Logging","TIMER",TempMessage);
    
    sprintf(TempMessage,"%d",debug_vif);
    WriteCfgString("Logging","VIF",TempMessage);
    
    sprintf(TempMessage,"%d",debug_thread);
    WriteCfgString("Logging","THREAD",TempMessage);
    
    sprintf(TempMessage,"%d",debug_file);
    WriteCfgString("Logging","FILE",TempMessage);
}

/* Saves General vars in nSX2 Config File*/
void saveGeneralVars()
{   
    char TempMessage[100] ;
    sprintf(TempMessage,"%d",GS_VSync_Wait);
    WriteCfgString("General","GS_VSync",TempMessage); 
}

/* Load Log vars from nSX2 Config File*/
void loadLogVars()
{
    debug_gs = ReadCfgInt("Logging","GS",0) ; 
    debug_dma = ReadCfgInt("Logging","DMA",0) ; 
    debug_systemcall = ReadCfgInt("Logging","SYSCALL",0) ; 
    debug_gif = ReadCfgInt("Logging","GIF",0) ; 
    debug_intc = ReadCfgInt("Logging","INTC",0) ; 
    debug_timer = ReadCfgInt("Logging","TIMER",0) ; 
    debug_vif = ReadCfgInt("Logging","VIF",0) ; 
    debug_thread = ReadCfgInt("Logging","THREAD",0) ; 
    debug_file = ReadCfgInt("Logging","FILE",0) ; 
}

/* Load General vars from nSX2 Config File*/
void loadGeneralVars()
{
    GS_VSync_Wait = ReadCfgInt("General","GS_VSync",0) ; 
}


/* Load all vars from nSX2 Config File*/
void loadSettings()
{
    loadLogVars() ;
    loadGeneralVars() ;
}


BOOL CALLBACK LogSheetProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
    switch(Message) {
    case WM_INITDIALOG:
         if  (debug_gs) {
                  SendMessage(GetDlgItem(hwnd,IDC_LOG_GS),BM_SETCHECK, BST_CHECKED,0);
                }
         if  (debug_dma) {
                  SendMessage(GetDlgItem(hwnd,IDC_LOG_DMA),BM_SETCHECK, BST_CHECKED,0);
                }
         if  (debug_systemcall) {
                  SendMessage(GetDlgItem(hwnd,IDC_LOG_SYSCALL),BM_SETCHECK, BST_CHECKED,0);
                }
         if  (debug_gif) {
                  SendMessage(GetDlgItem(hwnd,IDC_LOG_GIF),BM_SETCHECK, BST_CHECKED,0);
                }
         if  (debug_intc) {
                  SendMessage(GetDlgItem(hwnd,IDC_LOG_INTC),BM_SETCHECK, BST_CHECKED,0);
                }
         if  (debug_timer) {
                  SendMessage(GetDlgItem(hwnd,IDC_LOG_TIMER),BM_SETCHECK, BST_CHECKED,0);
                }
         if  (debug_vif) {
                  SendMessage(GetDlgItem(hwnd,IDC_LOG_VIF),BM_SETCHECK, BST_CHECKED,0);
                } 
         if  (debug_thread) {
                  SendMessage(GetDlgItem(hwnd,IDC_LOG_THREAD),BM_SETCHECK, BST_CHECKED,0);
                } 
         if  (debug_file)  {
                  SendMessage(GetDlgItem(hwnd,IDC_LOG_FILE),BM_SETCHECK, BST_CHECKED,0);
                }       
                                       
         TranslateLogSheet(hwnd) ;                           
         return TRUE;
    case WM_NOTIFY:
               if (((NMHDR FAR *) lParam)->code == PSN_APPLY)  {
                              debug_gs = SendDlgItemMessage(hwnd,IDC_LOG_GS,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;
                              debug_dma = SendDlgItemMessage(hwnd,IDC_LOG_DMA,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;
                              debug_systemcall = SendDlgItemMessage(hwnd,IDC_LOG_SYSCALL,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;
                              debug_gif = SendDlgItemMessage(hwnd,IDC_LOG_GIF,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;
                              debug_intc = SendDlgItemMessage(hwnd,IDC_LOG_INTC,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;
                              debug_timer = SendDlgItemMessage(hwnd,IDC_LOG_TIMER,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;
                              debug_vif = SendDlgItemMessage(hwnd,IDC_LOG_VIF,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;
                              debug_thread = SendDlgItemMessage(hwnd,IDC_LOG_THREAD,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;
                              debug_file = SendDlgItemMessage(hwnd,IDC_LOG_FILE,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;
                             
                              saveLogVars();
                              }
            break;                     
     default:
            return FALSE;       
     }
     return TRUE;            
}


////////////////////////////////////////////////////////////////////////////////

BOOL CALLBACK GeneralSheetProc(HWND hwnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
    switch(Message) {
    case WM_INITDIALOG:
         if  (GS_VSync_Wait) {
                  SendMessage(GetDlgItem(hwnd,IDC_GENERAL_GSVSYNC),BM_SETCHECK, BST_CHECKED,0);
                }
         TranslateGeneralSheet(hwnd) ;         
         return TRUE;
    case WM_NOTIFY: 
            if (((NMHDR FAR *) lParam)->code == PSN_APPLY)  {
              GS_VSync_Wait = SendDlgItemMessage(hwnd,IDC_GENERAL_GSVSYNC,BM_GETCHECK , 0,0) == BST_CHECKED?TRUE:FALSE;          
              saveGeneralVars();            
            }    
    default:
         return FALSE; 
    }
    return TRUE;                    
}


////////////////////////////////////////////////////////////////////////////////
/* Main Sheets */

void ChangeSettings(HWND hwndOwner) {
    PROPSHEETPAGE psp[2];
    PROPSHEETHEADER psh;
	char LogStr[200], titleStr[200], GeneralStr[200];
    
    psp[0].dwSize = sizeof(PROPSHEETPAGE);
    psp[0].dwFlags = PSP_USETITLE;
    psp[0].hInstance = App_hInstance;
    psp[0].pszTemplate = MAKEINTRESOURCE(IDD_LOGOPTIONS);
    psp[0].pfnDlgProc = LogSheetProc;
	TranslateDefault("Log options","Log options",LogStr);
	psp[0].pszTitle = LogStr;
    psp[0].lParam = 0;
    psp[0].pfnCallback = NULL;
    
    psp[1].dwSize = sizeof(PROPSHEETPAGE);
    psp[1].dwFlags = PSP_USETITLE;
    psp[1].hInstance = App_hInstance;
    psp[1].pszTemplate = MAKEINTRESOURCE(IDD_GENERALOPTIONS);
    psp[1].pfnDlgProc = GeneralSheetProc;
	TranslateDefault("General","General",GeneralStr);
	psp[1].pszTitle = GeneralStr;
    psp[1].lParam = 0;
    psp[1].pfnCallback = NULL;

    psh.dwSize = sizeof(PROPSHEETHEADER);
    psh.dwFlags = PSH_PROPSHEETPAGE | PSH_NOAPPLYNOW;
    psh.hwndParent = hwndOwner;
    psh.hInstance = App_hInstance;
    TranslateDefault("Settings","Settings",titleStr);
    psh.pszCaption = (LPSTR) titleStr;
    psh.nPages = sizeof(psp) / sizeof(PROPSHEETPAGE);
    psh.nStartPage = 0;
    psh.ppsp = (LPCPROPSHEETPAGE) &psp;
    psh.pfnCallback = NULL;

	PropertySheet(&psh);
	return;
}
