/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>


#include <time.h>



#include "nSX2.h"

#include "PS2Edefs.h"
#include "_Debuger/debuger.h"

#include "pluggins_main.h"

#include "sif.h"
#include "r5900_main.h"
#include "r5900_core.h"
#include "r5900_core_table.h"
#include "timers.h"
#include "bios_syscalls.h"
#include "bios_syscalls_table.h"
#include "VU.h"
#include "bios.h"

#include "GUI_MainWindowCallbacks.h"

#include "main.h"	//for the Status Bar

extern char * debug_r5900_reg [34];

float ComputeInstructionsPerSec( void  );

_u8		debug_op;
_u8		debug_regs;



_u8		debug_thread ;
_u8		debug_systemcall;	

_u8		debug_gif;
_u8		debug_intc;
_u8		debug_timer;
_u8		debug_vif;
_u8		debug_gs;
_u8		debug_dma;
_u8     debug_file; 

_u8     GS_VSync_Wait;

/*************************************/

/* nSX2 Variables Declarations */
Mips_CPU 		MipsCpu;
Bios_Hw  		Ps2Bios;
VPU0			VectorUnit0;	
VPU1			VectorUnit1;

/* Syscall Dynamic Functions */
void (*syscall_fct_ns[256])(void);



/* nSX2 Functions Declarations */
void 	StopEmulation();


void 	PrintToOpcodeOutput 			( char * ftext );
void    PrintToDmaOutput                ( char * ftext ); 
void    PrintToBIOSOutput               ( char * ftext );
void 	PrintToRegistersOutput 			( char * ftext );
void 	PrintToCOP1RegistersOutput 		( char * ftext );
void    PrintExtraStuff                 ( void );




/* WIN 32 Global */
HANDLE hThread	= NULL; 
HWND  dwThrdParam;





/*************************************************************************/
/*					Emulator Initialisation Function		 	 		*/
/*	Open the Log File													*/
/*	Load the Config ( in .ini file )									*/
/* 	Load Plugins ( GS - Input - SIF and Debuger							*/
/*																		*/
/*	Note :	Plugin check is missing										*/
/*			Plugin version is missing									*/
/*																		*/
/*************************************************************************/

int nSX2_Init( void )
{ 


	char tempText[100];
	_u8		Result = 0;

	_u32	LibInfoVersion;
	_u8		LibVersion;
	_u8		LibRevision;	
	_u8		LibBuild;
	_u32    PADqueryResult;


	
	char* path;

	GetCurrentDirectory( MAX_PATH, cCurDir );
 
	/* Read config File
		If errors happend, no problem this functions load default config
	*/
	nSX2LoadConfig();							

	
	/*Load GS Pluggin Function */
    Result = LoadGsPluggin(GSPluginToUse);

	/*Load Input Pluggin Function */
	Result = LoadInputPluggin( InputPluginToUse );


	/*Load SIF Pluggin Functions */
	Result = LoadSifPluggin( SifPluginToUse );

   if ( Result == 0 ) {
		/*Bind the RPC Function of the SIF Plugins */
		Result = BindSifPluggin();
		
	}
	
	/*Load Debug Pluggin Functions */
	Result = LoadDEBUGPluggin( DBGPluginToUse ) ;


	/* Debug Var */	
	debug_op			=	0;
	debug_regs			=	0;
	
	return 0; 
}




/*************************************************************************/
/*					PS2 Initialisation Function				 	 		*/
/*	Read Elf Header														*/
/*	Alloc Memory														*/
/* 	Load elf in Memory													*/	
/*	GS Plugin Initialisation											*/
/*	Pad Plugin Initialisation											*/
/*	PS2 Device Initilisation											*/
/*																		*/
/*	Note :	Initialisation check is missing								*/
/*																		*/
/*************************************************************************/
int	GSAlreadyLoad = FALSE;	//


int Emulation_Init  ( HWND hwnd , _u8 LoadType ) {
	char 		tempText[100];		//for displaying text to the widget

	static int	IsInitOK = 0;		//
	
	int 		result;				//fonction return
	int 		boucle;				//loop
	char 		GSMainTitle[256];	//main windows title	



#ifdef EMU_LOG
	EMU_LOG("nSX2 %1.1d.%2.2d.%1.1d",  nSX2_VERSION_MAJ, nSX2_VERSION_MIN,nSX2_VERSION_BUILD);
	EMU_LOG("|  web  : http://nsx2.emulation64.com");
	EMU_LOG("|_ mail : atreides.muaddib@caramail.com");
#endif		
 
	switch ( LoadType ) {
		case LOAD_NOTHING 	:
		case LOAD_CDVD 		:		break;
		
		case LOAD_ELF_FILE 	:
			/*Open the elf File */
			result = ElfOpenFile(ElfToExecute);	
			if ( result == FALSE )  IsInitOK++;
		break;
		
		case LOAD_BIOS 		:
			/*Open the BIOS File */
			result = BiosOpenFile( "bios/scph10000.bin" ) ;		
			if ( result == FALSE )  IsInitOK++;
		break;
	}
		
	/*Various Memory Allocation */
	FreeAllMemory();		/* if memory was already allocated, free it */
    result = MipsMemoryAlloc();							if ( result == FALSE )  IsInitOK++;
	result = MipsSpecialAreaMemoryAlloc();				if ( result == FALSE )  IsInitOK++;
	result = MipsGSAreaMemoryAlloc();					if ( result == FALSE )  IsInitOK++;
	result = MipsBiosMemoryAlloc();						if ( result == FALSE )  IsInitOK++;
	result = MipsSPRMemoryAlloc();						if ( result == FALSE )  IsInitOK++;

	
		
	/*If something wrong with initialisation*/
	if ( IsInitOK != 0 ) {
#ifdef EMU_LOG
		EMU_LOG("FATAL ERROR while emulator initialisation....Sorry Exiting");	
#endif
		FreeAllMemory();		
		SetInterpreterRunning ( FALSE ) ;
		return FALSE;
	}


	/* Load ... */
	switch ( LoadType ) {

		case LOAD_CDVD :		
			/*... elf from the CDVD Plugin ( through the SIF Plugin ) */
			/* if Loading OK ...*/
			if ( SIFRun_DVD() == TRUE ) MipsCpu.EntryPointAdd = SIFGetElfAdd();	// Get the EntryPoint
			/* ...else Stop emulation */
			else {
#ifdef EMU_LOG
				EMU_LOG("FATAL ERROR while CDVD initialisation....Sorry Exiting");	
#endif
				FreeAllMemory();		
				SetInterpreterRunning ( FALSE ) ;
				return FALSE;
			}
		break;

		case LOAD_ELF_FILE :
			/* ... elf from a file */
			ElfLoadHeader();		/*Read the header of the Elf File*/
			CopyProgramToMemory() ;	/*Copy the program to memory according the elf header*/
			CopySectionToMemory( );	/*Copy data to memory according the elf header*/
			ElfFreeHeader();		/*Free Memory Header */

		break;

		case LOAD_BIOS :		
			/* ... bios into memory*/
			CopyBiosToMemory ( );
			BiosCloseFile();
		break;

		case LOAD_NOTHING :		break;

	}


	/* Init all Plugins */	
	IsInitOK = PLUGINS_InitAll();
	if ( IsInitOK == FALSE ) {
#ifdef EMU_LOG
		EMU_LOG("FATAL ERROR while Plugins initialisation....Sorry Exiting");	
#endif
		FreeAllMemory();		
		SetInterpreterRunning(  FALSE ); 
		return FALSE;
	}

	
	/*Various PS2 Initialisation */
	DMAC_Init			( );
	INTC_Init			( );
	VIF_Init			( );
	GIF_init			( );
	TIMERS_Init			( );
	R5900_Core_init    	( );
	R5900_Cop0_init     ( );
	R5900_Cop1_init		( );
	VU_init				( );


	/* Need To refresh at least 1 time the display */
	Ps2Bios.NeedUpdateScreen 		= TRUE;
			
	/* Don't need to read Pad Status until I need it */
	Ps2Bios.NeedUpdatePad 			= FALSE;
			
	/* Timer Init */
	Ps2Bios.NbOfInstructionsPerSec	= 10000000;


   	Ps2Bios.AccesToGS = FALSE;

	/* Bios / Kernel Init */
	for ( result = 0; result < BIOS_MAX_SEMAPHORE; result++ ) {
		Ps2Bios.SemaState[result] = BIOS_SEMA_ROOMFREE ;
		InitQueue( Ps2Bios.Sema[ result ].WaitQueue);
	}

	/* PS2 Threads Init */
	THREADS_Init();
	
	/* 	Syscall Dynamic Table 	*/
	memcpy(syscall_fct_ns,syscall_fct, sizeof(void *) * 256 );
	memset( Ps2Bios.SysCallUserDef, (_u32)0xFFFFFFFF, sizeof( _u32 ) * 256 );

	/* Pad Initialisation */
	Ps2Bios.PadData[0][0] = NULL;
	Ps2Bios.PadData[0][1] = NULL;
	Ps2Bios.PadData[1][0] = NULL;
	Ps2Bios.PadData[1][1] = NULL;

	Ps2Bios.XPadData[0][0] = NULL;
	Ps2Bios.XPadData[0][1] = NULL;
	Ps2Bios.XPadData[1][0] = NULL;
	Ps2Bios.XPadData[1][1] = NULL;
		
	Ps2Bios.PadState[0] = NULL;
	Ps2Bios.PadState[1] = NULL;
		
	
	GUI_InitDebuger( hwnd );

	switch ( LoadType ) {
		case LOAD_CDVD :
		case LOAD_ELF_FILE :	MipsCpu.halt = 0;		break;
		case LOAD_NOTHING :		MipsCpu.halt = 1;		break;
		case LOAD_BIOS :								
			MipsCpu.halt 	= 1;
			MipsCpu.PC		= (_u32)0xBFC00000;
		break;
	}

	/* The Interprter is running */
	SetInterpreterRunning ( TRUE ) ;  

	IsInitOK = 0;

#ifdef EMU_LOG
	EMU_LOG("Start Interpreter....");
#endif		
	
	return TRUE;
}




/*************************************************************************/
/*					PS2 Interpreter 					 	 			*/
/*																		*/
/*	PS2 Initialisation													*/
/*	Main Loop															*/
/*																		*/
/*	Note :																*/
/*																		*/
/*************************************************************************/
DWORD WINAPI Emulation_Interp( LPVOID lpParam )  
{ 
 	_u16	boucle;
    MSG 	Msg;



	/* 	Emulation Initialisation 
		LoadElfFile is a global variable ( not very clean )	
	*/ 
	Emulation_Init  ( lpParam, LoadElfFile );

	/* While the Interpreter is running... */
	while ( GetInterpreterRunning() == TRUE ) {

		/* Pad Update, Input Keyboard, GS & Thread Sheduler */	
		/* 	only Update when the main CPU is runnnig .
			This is why, input key don't work when the emulation is halted
		*/
		if ( MipsCpu.halt != 1 ) {
			TIMERS_Update();		// Update Timer and Main Clock 
		} else {
			KeyEvent( (*PAD1keyEvent)() );
		}

		switch( MipsCpu.halt ) {

			/* the CPU is running */
			case 0: 
				for ( boucle = 0 ; boucle < 1000 ; boucle ++ ) { 
    				MipsUpdateCpu(); 
        			if ( MipsCpu.halt != 0) break;
       			} //Execute a Block of opcodes
			break;

			/* case1 : the CPU is halted */

			/* the CPU execute ONE opcode and Stop */
			case 2:
				MipsUpdateCpu();	//Execute the current opcode
				MipsCpu.halt = 1;	//And Stop
			break;
		
		}
		
		/* Need Comments ? */
		ComputeInstructionsPerSec();

	} /* ... end of while  Interpreter is running */

	/* Stop the Emulation */
	StopEmulation();


	return 0;
} 






/*************************************************************************/
/*					PS2 Stop Interpreter Function			 	 		*/
/*																		*/
/*	Shutdown Plugins													*/
/*	Close Open File														*/
/*	Free All Memory														*/
/*																		*/
/*																		*/
/*	Note :																*/
/*																		*/
/*************************************************************************/
void StopEmulation()
{
 	char tempText[200];  	



	/*Free All Emu Memory*/
	FreeAllMemory();

	/* PS2 Emulation Threads Shutdown */
	THREADS_Shutdown();

	/* Close GS & Input Plugins */
	PLUGINS_ShutdownAll();


	GUI_ShutdownDebuger( NULL );

#ifdef EMU_LOG
	EMU_LOG("...Stop Interpreter");
	EMU_LOG("Stop at PC = 0x%8.8llx after %ld instructions",(_u32)MipsCpu.PC, (_u32)MipsCpu.Nb_instructions);
#endif		

	SetStatusText("Emulation Stopped");

}


/*************************************************************************/
/*					Emulation Thread Creation							*/
/*																		*/
/*																		*/
/*************************************************************************/
void Init_Emulation_Interp ( HWND hwnd ) {

	DWORD dwThreadId; 

 	dwThrdParam = hwnd;

	/* If an .elf is already running... */
	if ( MipsCpu.InterpRunning 	== TRUE ) {

		GUI_SendStopMainLoop( TRUE );
	
	}

	hThread = CreateThread( 
        NULL,                        // no security attributes 
        0,                           // use default stack size  
        Emulation_Interp, 	 		// thread function 

        &dwThrdParam,                // argument to thread function 
        0,                           // use default creation flags 
        &dwThreadId);                // returns the thread identifier 
	

	if (hThread == NULL) 
    	  MessageBox(hwnd, "Could not Create Emu Thread!", "Error", MB_OK | MB_ICONERROR); 




}


/********************************************************************************/
/*  R5900 SPEED           														*/
/********************************************************************************/

float ComputeInstructionsPerSec( void  ) {

        static int	PreviousTime = 0.0;
        static _u32	PreviousNbInstruction = 0;
        int	CurrentTime;
        float	InstPerSec;
        
        char temp_string [100];

        CurrentTime =  timeGetTime(); //gets time
        
		if ( CurrentTime < PreviousTime ) PreviousTime = 0.0;
        
		if ( ( CurrentTime - PreviousTime ) >= 1000 ) {

			Ps2Bios.NbOfInstructionsPerSec = (_u64)(( MipsCpu.Nb_instructions - PreviousNbInstruction ) / (float)( CurrentTime - PreviousTime ) ) ;

			InstPerSec = (float)(Ps2Bios.NbOfInstructionsPerSec * 1000 );
			InstPerSec = (float)(InstPerSec / 1000000 );

				if ( InstPerSec == (float)0.0 ) {
					sprintf(temp_string,"Emulation Stopped");
				} else {
					sprintf(temp_string,"CPU Speed:%10.6f MI / s\n",InstPerSec);
				}

			SetStatusText(temp_string);
	
			PreviousNbInstruction = MipsCpu.Nb_instructions;
			PreviousTime = 	CurrentTime;
		}

}











