/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#ifndef _INTRC_H
#define _INTRC_H

/* Define or Undefine if you want INTC LOG */
#define	INTC_LOG	INTC_Log
//#undef	INTC_LOG

/********************************************************************************/
/*						INTC DEFINES											*/
/********************************************************************************/

#define	INTC_STAT				0x1000F000	//Interrupt status
#define	INTC_MASK				0x1000F010  //Interrupt mask
#define	STD_OUT_STATUS			0x1000F130  //STD-OUT STATUS???
#define	STD_OUT_DATA			0x1000F180  //      ???


#define	BIOS_MAX_DMA_HANDLER	10
#define	BIOS_MAX_INTC_HANDLER	11



/********************************************************************************/
/*						INT0 DEFINES											*/
/********************************************************************************/

#define INTC_GS  			0
#define INTC_SBUS  			1
#define	INTC_VBLANK_S		2
#define	INTC_VBLANK_E		3
#define INTC_VIF0  			4
#define INTC_VIF1  			5
#define INTC_VU0  			6
#define INTC_VU1  			7
#define INTC_IPU  			8
#define INTC_TIM0  			9
#define INTC_TIM1  			10



/********************************************************************************/
/*						INT HANDLER STRUCTURES									*/
/********************************************************************************/

typedef struct _LinkedAddress {

	_u64	Add;					//Address of the Handler
	_u32	ID;						//ID Handler
	struct	_LinkedAddress * next;	

}LinkedAddress;


typedef struct	_IntCHandler {
	
	_u32		Status;			//IntC Handler Status : Enabled  / Disabled

	_u32		Numero_ID;		//Members to find a free ID
	_u32		Nb_handler;


	LinkedAddress * HandlerAdd;
		
} IntCHandler;


typedef struct	_DmaHandler {
	
	_u32		Status;			//Dma Handler Status : Enabled  / Disabled
	_u32		Numero_ID;		//Members to find a free ID
	_u32		Nb_handler;


	LinkedAddress * HandlerAdd;
} DmaHandler;



/********************************************************************************/
/*						INTC REGISTERS STRUCTURE								*/
/********************************************************************************/


typedef struct _InteruptControllerRegisters {

	_u32	stat;			// 0x00  -  INTC STATUS 
	_u32	null0[3];		// 0x04 0x08 0x0C
	_u32	mask;			// 0x10  -  INTC INTERRUPT MASK

} InteruptControllerRegisters;




/********************************************************************************/
/*						INTC  STRUCTURE											*/
/********************************************************************************/

#define		NO_EXECUTING_HANDLER	0
#define		EXECUTING_HANDLER		1
#define		NEED_UPDATE_SCHEDULER	2


typedef struct _INTController {

	/* Registers */
	InteruptControllerRegisters *Reg;

	/* INT Handler */
	IntCHandler 				IntByCause	[BIOS_MAX_INTC_HANDLER];
	

	/* Dma Handler */
	DmaHandler					IntByDMA	[BIOS_MAX_DMA_HANDLER];


	_u8 						IntrEnable;

	_u8							ExecutingHandler;

}INTController;


extern INTController 		IntC;


/********************************************************************************/
/*						INTC  FUNCTIONS											*/
/********************************************************************************/

/* Main */
void 	INTC_Log 		( char * ftext  , ...   );
void	INTC_Init		( void );
void	INTC_Shutdown 	( void );

void	INTC_Read 	( _u64 fPhysicalAddress, _u128 * fVal, 	_u8 fDataType );
void	INTC_Write 	( _u64 fPhysicalAddress, _u128 fVal, 	_u8 fDataType );

__inline _u32 	INTC_ExecInterrupt	( _u32 HandlerAdress );


/* INT0 */
void 	INTC_RaiseIntByCause	( _u8 fNumber );
void 	INTC_TriggerINT0		( _u8 IntrNumber );


/* INT1 */
void 	INTC_RaiseIntByDMA 		( _u8 fNumber );
void 	INTC_TriggerINT1		( _u8 Channel );


/* Execute Sub Function */
#define	INTC_SUBF_SAVE_REG	1
#define	INTC_SUBF_SET_A0	2
#define	INTC_SUBF_SET_A1	4
#define	INTC_SUBF_SET_A2	8

_u32 INTC_ExecSubFunc( _u32 fAdress ,_u32 fOptions , _u32 fA0 , _u32 fA1, _u32 fA2);


#endif














