/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#include <stdlib.h>
#include <stdio.h>

#include    "../nsx2.h"
#include	"../nSX2_type.h"

#include 	"../bios.h"

#include	"LinkedList.h"



 


/* Create a new List Item with the data 
	return the NewItem Adress if succes
	else return NULL
*/
LLThread *	LLThread_CreateItem ( PS2_Thread * fData ) {

	LLThread * LL_NewItem;

	LL_NewItem = ( LLThread * ) malloc ( sizeof( LLThread ) );
	
	if ( LL_NewItem == NULL ) return NULL;
	else {

		LL_NewItem->next 		= NULL;
		LL_NewItem->previous	= NULL;

		LL_NewItem->nextALL 	= NULL;
		LL_NewItem->previousALL	= NULL;


		/* DATA COPY HERE */
		LL_NewItem->thread = ( PS2_Thread * ) malloc ( sizeof ( PS2_Thread ) );
		memcpy( LL_NewItem->thread, fData, sizeof ( PS2_Thread ));

		return LL_NewItem;
	}

}
/*
	Free Memory

*/
void	LLThread_Free( LLThread  ** fList ) {

	LLThread  * NextItem;

	if ( (*fList) == NULL ) return;

	while ( (*fList) != NULL ) {

		NextItem = (*fList)->next;
	
		/* FREE DATA HERE IS NEEDED */
		free ( (*fList)->thread );
		free ( (*fList)->Stack );

		free( (*fList) );
		(*fList) = NULL;

		(*fList) = NextItem;

	}

}

void	LLThread_Free_OneThread ( LLThread * fItem ) {

	LLThread * WorkItem;
	LLThread * PreviousItem;	
	
	PreviousItem	= Ps2Bios.PS2_Sched.All_Threads;
	WorkItem 		= Ps2Bios.PS2_Sched.All_Threads;


	if ( fItem == WorkItem ) {

		Ps2Bios.PS2_Sched.All_Threads = WorkItem->nextALL;

		free ( WorkItem->thread );
		free ( WorkItem->Stack );
		free ( WorkItem );


	} else {

		while ( WorkItem != NULL ) {
			if ( WorkItem == fItem ) {
	
				PreviousItem->nextALL = WorkItem->nextALL;

				free ( WorkItem->thread );
				free ( WorkItem->Stack );
				free ( WorkItem );

				break;

			}
			PreviousItem 	= WorkItem;
			WorkItem 		= WorkItem->nextALL;
		}
	}

}

/*
	Find a Item in the List
	if found return the Item
	else return NULL
*/

LLThread *	LLThread_Find ( LLThread * fList , _u32 fID )	{
	LLThread * WorkItem;	

	if ( fList == NULL ) return NULL;

	WorkItem = fList;

	while ( WorkItem != NULL ) {
		if ( WorkItem->ID == fID ) return WorkItem;
		WorkItem = WorkItem->nextALL;
	}

	return NULL;

}

/* End of Specific Function */
/******************************************************************************/


/* Initialisation of the List */
void	LLThread_Init( LLThread * fList ) {
	fList->next			= NULL;

}


/*
	Add Item at the End of the List
*/
void	LLThread_AddEnd( LLThread ** fList, LLThread * fItem ) {

	LLThread * WorkItem;

	if ( fItem == NULL ) return;

	WorkItem = 	(*fList);



	if ( WorkItem != NULL ) {
	
		while ( WorkItem->next != NULL ) 	WorkItem = WorkItem->next;

		WorkItem->next	= fItem;


	} else 	(*fList) = fItem;

	fItem->next = NULL;

}

void	LLThread_AddEnd_ALL( LLThread ** fList, LLThread * fItem ) {

	LLThread * WorkItem;

	if ( fItem == NULL ) return;

	WorkItem = 	(*fList);

	if ( WorkItem != NULL ) {
	
		while ( WorkItem->nextALL != NULL ) 	WorkItem = WorkItem->nextALL;

		WorkItem->nextALL	= fItem;


	} else 	(*fList) = fItem;

}


/*
	At Item at the Begining of the List
*/
void	LLThread_AddFirst ( LLThread ** fList, LLThread * fItem ) {

	if ( fItem == NULL ) return;

	if ( (*fList) == NULL ) (*fList) = fItem;
	else {

		fItem->next		= (*fList);

		(*fList)		= fItem;
	}

}

/*
	Remove First Item of the List
	if Destroy item return NULL
	else return the Item

*/
LLThread *	LLThread_RemoveFirst ( LLThread ** fList, _u8 fDestroy) {

	LLThread * NextItem;
	LLThread * ReturnItem;	

	if ( (*fList) == NULL ) return;

	NextItem = (*fList)->next;

	if ( fDestroy == LL_WITH_FREE )	free ( (*fList) );
	else {
		ReturnItem = *fList;
		ReturnItem->next		= NULL;

	}


	(*fList) = NextItem;
//	(*fList)->previous = NULL;

	if ( fDestroy == LL_WITH_FREE ) return NULL;
	else return ReturnItem;

}

/*
	Remove Last Item of the List
	if Destroy item return NULL
	else return the Item

*/
LLThread *	LLThread_RemoveEnd ( LLThread ** fList , _u8 fDestroy) {

	LLThread * ReturnItem;	
	LLThread * WorkItem;
	LLThread * PreviousItem;

	if ( (*fList) == NULL ) return;

	WorkItem = (*fList);
	
	while ( WorkItem->next != NULL ) {
		PreviousItem = WorkItem;
		WorkItem = WorkItem->next;	

	}

	if ( fDestroy == LL_WITH_FREE )	free ( WorkItem );
	else {
		ReturnItem = WorkItem;
		ReturnItem->next		= NULL;

	}

	PreviousItem->next = NULL;

	if ( fDestroy == LL_WITH_FREE ) return NULL;
	else return ReturnItem;


}

/*
	Remove Specificied Item of the List
	if Destroy item return NULL
	else return the Item

*/

LLThread *	LLThread_Remove ( LLThread ** fList, LLThread * fItem , _u8 fDestroy) {
	LLThread * ReturnItem = NULL;	
	LLThread * WorkItem;
	LLThread * PreviousItem;

	if ( (*fList) == NULL ) return;

	WorkItem = (*fList);

	if ( WorkItem == fItem ) {

		

		if ( fDestroy == LL_WITH_FREE )	{

			LLThread_Free_OneThread (WorkItem );
		}
		else {
			(*fList) 			= WorkItem->next;
			ReturnItem = WorkItem;
			ReturnItem->next		= NULL;

		}

	} else {

		PreviousItem = WorkItem;

		while ( WorkItem != NULL ) {
			if ( WorkItem == fItem ) {
				
						
				if ( fDestroy == LL_WITH_FREE )	LLThread_Free_OneThread ( WorkItem );
				else {
					PreviousItem->next 	= WorkItem->next;
					ReturnItem 			= WorkItem;
					ReturnItem->next	= NULL;
				
				}

			}

			
			PreviousItem = WorkItem;
			if ( fDestroy == LL_NO_FREE ) 	WorkItem = WorkItem->next;
			else WorkItem = WorkItem->nextALL;


		
		}
	}

	if ( fDestroy == LL_WITH_FREE ) return NULL;
	else return ReturnItem;


}


/********************************************************************************/

char LL_DEBUG_STRING[2048];

/* Print the content of the List */
char * LLThread_PrintList( char * fTitle, LLThread * fList ) {
	LLThread * WorkItem;
	

	WorkItem = fList;

	sprintf(LL_DEBUG_STRING,"%s ",fTitle);
	

	while ( WorkItem != NULL ) {
		sprintf(LL_DEBUG_STRING,"%s %d(%d) -> ",LL_DEBUG_STRING,WorkItem->ID,WorkItem->thread->currentPriority);
		WorkItem = WorkItem->next;
	}
	
	sprintf(LL_DEBUG_STRING,"%s NULL \n",LL_DEBUG_STRING);

	return LL_DEBUG_STRING;
}

char * LLThread_PrintListAll( char * fTitle, LLThread * fList ) {
	LLThread * WorkItem;
	

	WorkItem = fList;

	sprintf(LL_DEBUG_STRING,"%s ",fTitle);
	

	while ( WorkItem != NULL ) {
		sprintf(LL_DEBUG_STRING,"%s %d -> ",LL_DEBUG_STRING,WorkItem->ID);
		WorkItem = WorkItem->nextALL;
	}
	
	sprintf(LL_DEBUG_STRING,"%s NULL \n",LL_DEBUG_STRING);

	return LL_DEBUG_STRING;
}




