/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <errno.h>
#include <string.h>

#include "nSX2.h"
#include "Loader.h"
#include "Loader_elf.h"

FILE * ElfFile;
elf_header ico_elf_hdr;
elf_program_header * ico_elf_prg_hdr ;
elf_section_header * ico_elf_sec_hdr ;
long ElfFileSize;

FILE * BiosFile;

 

long FileSizeOf( FILE *fp )
{
    long int save_pos;
    long size_of_file;

    save_pos = ftell( fp );
    fseek( fp, 0L, SEEK_END );
    size_of_file = ftell( fp );
    fseek( fp, save_pos, SEEK_SET );
    return( size_of_file );
}


int CheckIfElf ( FILE *fp ) {

   	int ch;
   	int get = 0;
   	char tempText[100];

   	    fseek(fp,0,SEEK_SET);

        sprintf(tempText,"");
        
        while ((ch=fgetc(fp)) != EOF) {
        
                get++;

                if (get >= 5) { break; }
                
                if (get == 2 || get == 3 || get == 4) { sprintf(tempText,"%s%c",tempText,ch); }
                
        }
        
        if (!strcmp(tempText,"ELF")) {
            return TRUE; // file is an .elf file
        }
        else {
        	return FALSE; // file is not an .elf file
#ifdef EMU_LOG
		    EMU_LOG("[LOADER][ELF]file isn't a valid elf file");
#endif

        }

}

_u8 ElfOpenFile( char * fFileName ) {

	_u8 	Result;

	ElfFile = fopen( fFileName, "rb" );
	
	if( ElfFile != NULL )  {
		ElfFileSize = FileSizeOf( ElfFile ) ;

		#ifdef EMU_LOG
		    EMU_LOG("[LOADER][ELF]Open : %s ( size : %5.3f Ko)",fFileName , ( (float) ElfFileSize  / 1024));
        #endif
		
	}
	else	{

	    #ifdef EMU_LOG
			    EMU_LOG("[LOADER][ELF]Error opening :%s: %s", fFileName ,strerror( errno ));
		#endif

		return FALSE;
	}
	
	// Check if file ins an .elf file
	if ( CheckIfElf (ElfFile) ) { 
		Result = TRUE; 
	} else { 
		Result = FALSE; 
		ElfCloseFile();
	}
	
	return Result;	
}


void ElfCloseFile( void ) {

#ifdef EMU_LOG
	EMU_LOG("[LOADER][ELF]Close");
#endif
		
	if ( ElfFile != NULL ) {

		fclose( ElfFile );
		ElfFile = NULL;

	}
/*
	if ( ico_elf_prg_hdr != NULL ) 	{	free(ico_elf_prg_hdr);		ico_elf_prg_hdr 	= NULL; }
	if ( ico_elf_sec_hdr != NULL ) 	{	free(ico_elf_sec_hdr);		ico_elf_sec_hdr 	= NULL; }
*/

}


/*

MEMORY ALLOC ISN'T FREE


*/
_u8 ElfLoadHeader() {
	
	_u8  Index;
	_u8 Result;
	char tempText[100];
		
	Result = 0;

	fseek( ElfFile, 0L, SEEK_SET );

	if ( fread( &ico_elf_hdr , sizeof(struct  _elf_header ), 1, ElfFile ) != 0 ) {
		MipsCpu.EntryPointAdd = ico_elf_hdr.entry_point_add;
	}
	else Result++;

	
	fseek( ElfFile, ico_elf_hdr.start_prg_hdr, SEEK_SET );


	if((ico_elf_hdr.num_of_prg_hdr * ico_elf_hdr.prg_hdr_entry_size) != 0) {

		ico_elf_prg_hdr = (struct _elf_program_header *) malloc(ico_elf_hdr.num_of_prg_hdr * ico_elf_hdr.prg_hdr_entry_size);

		fread((_u8 *)ico_elf_prg_hdr, 1, ico_elf_hdr.num_of_prg_hdr * ico_elf_hdr.prg_hdr_entry_size, ElfFile);
	}
	else
		ico_elf_prg_hdr = NULL;




	
	fseek( ElfFile, 0L, SEEK_SET );


	
	fseek( ElfFile,  ico_elf_hdr.start_sct_hdr, SEEK_SET );



	if((ico_elf_hdr.num_of_sec_hdr * ico_elf_hdr.sec_hdr_entry_size) != 0) {

		ico_elf_sec_hdr = (struct _elf_section_header *) malloc(ico_elf_hdr.num_of_sec_hdr * ico_elf_hdr.sec_hdr_entry_size);

		fread((_u8 *)ico_elf_sec_hdr, 1, ico_elf_hdr.num_of_sec_hdr * ico_elf_hdr.sec_hdr_entry_size, ElfFile);
	}
	else
		ico_elf_sec_hdr = NULL;


	
#ifdef EMU_LOG
	EMU_LOG("[LOADER][ELF][HEADER]Read");
#endif
	

	return Result;
}

void 	ElfFreeHeader		(void) {

	if ( ico_elf_prg_hdr != NULL ) 	{	free(ico_elf_prg_hdr);		ico_elf_prg_hdr 	= NULL; }
	if ( ico_elf_sec_hdr != NULL ) 	{	free(ico_elf_sec_hdr);		ico_elf_sec_hdr 	= NULL; }

}





/*
THIS FUNCTION NEED TO BE REWRITE LIKE THE SECTION PART
*/
void CopyProgramToMemory( void ) {
	
	_u64 Addr;
	_u64 fIndex;
	_u64 Index2;
	_u64 ReadError;
	char c;
	char tempText[100];	
	int readOK;	


	_u32	FirstAdress = 0xFFFFFFFF;
	_u32	EndAdress = 0x0;

	_u16	NbPrgSectionLoaded = 0;
	
//	sprintf(tempText,"Init:Copy Program To Memory: ");
//	PrintToEmuOutput(tempText);	
	
	ReadError = 0;


	if ( ico_elf_prg_hdr != NULL ) {
	
		for ( Index2 = 0 ; Index2 <  ico_elf_hdr.num_of_prg_hdr ; Index2++ ) {

			switch ( ico_elf_prg_hdr[Index2].segment_type ) {

				case 0x1:
			
					


					Addr = ico_elf_prg_hdr[Index2].physic_add_of_segment;
		
					fseek( ElfFile,  ico_elf_prg_hdr[Index2].offs_from_start_to_prg_seg , SEEK_SET );	
	
					readOK = fread((_u8 *)(MipsMemory + Addr), 1, ico_elf_prg_hdr[Index2].num_of_bytes_in_file_ima, ElfFile);

					if ( (_u32)Addr < (_u32)FirstAdress )  FirstAdress = Addr ;
					if ( (_u32)(Addr+ico_elf_prg_hdr[Index2].num_of_bytes_in_file_ima) > EndAdress ) EndAdress = Addr ;

#ifdef LOADER_LOG
					LOADER_LOG("[LOADER][ELF]Prog %d : Loaded (size:%8.8lx)",Index2,ico_elf_prg_hdr[Index2].num_of_bytes_in_file_ima);
#endif
			
					NbPrgSectionLoaded++;

				break;


				default:
#ifdef LOADER_LOG
					LOADER_LOG("[LOADER][ELF]Prog %d : Unknown (size:%8.8lx)",Index2,ico_elf_prg_hdr[Index2].num_of_bytes_in_file_ima);
#endif
				break;
	
			}
		
		}
	
	}

	if (ReadError != 0 )	{
#ifdef EMU_LOG
		EMU_LOG("[LOADER][ELF][HEADER]Prog : Done with %lld errors",ReadError);
#endif

	}
	else	{
#ifdef EMU_LOG
	EMU_LOG("[LOADER][ELF][HEADER]Prog : Done ( %d sections Loaded )",NbPrgSectionLoaded);
#endif
	}
	


}


void CopySectionToMemory ( void ) {

	_u64 Addr = 0;
	_u32 index;
	_u64 size;
	_u64 ReadError;	

	_u8	* tempString;

	int readOK;	

	char c;
	char tempText[100];	
	_u16	NbPrgSectionLoaded = 0;	
		
//	sprintf(tempText,"Init:Copy Section To Memory: ");PrintToEmuOutput(tempText);	



	ReadError = 0;

	if ( ico_elf_prg_hdr != NULL ) {

		fseek( ElfFile,  ico_elf_hdr.start_sct_hdr, SEEK_SET );
		
		for ( index = 0 ; index <  ico_elf_hdr.num_of_sec_hdr ; index++ ) {

			

			switch ( ico_elf_sec_hdr[index].type ) {

			case SHT_NULL:

#ifdef LOADER_LOG
				LOADER_LOG("[LOADER][ELF]Section %d : NULL Loaded ( %d )",index,ico_elf_sec_hdr[index].size);
#endif


			break;


			case SHT_PROGBITS:
				Addr =  ico_elf_sec_hdr[index].sect_start_add & 0x1FFFFFFF;
				fseek( ElfFile,   ico_elf_sec_hdr[index].offset_from_start  , SEEK_SET );	


				readOK = fread((_u8 *)(MipsMemory + Addr), 1, ico_elf_sec_hdr[index].size, ElfFile);
			
				NbPrgSectionLoaded++;	

#ifdef LOADER_LOG
				LOADER_LOG("[LOADER][ELF]Section %d : PROGBITS Loaded ( %d )",index,ico_elf_sec_hdr[index].size);
#endif

 			break;


			case SHT_NOBITS:
#ifdef LOADER_LOG
				LOADER_LOG("[LOADER][ELF]Section %d : NOBITS Loaded ( %d )",index,ico_elf_sec_hdr[index].size);
#endif
			break;
			


			default:
#ifdef LOADER_LOG
				LOADER_LOG("[LOADER][ELF]Section %d : Unknown  %8.8lx (size:%8.8lx)",index,ico_elf_sec_hdr[index].type,ico_elf_sec_hdr[index].size);
#endif

			break;
			
 	
			}
		
		}

	}
						
#ifdef EMU_LOG
	EMU_LOG("[LOADER][ELF][HEADER]Section : Done ( %d sections Loaded )",NbPrgSectionLoaded);
#endif


}





/*
Debug Info
*/
void ElfPrintInfo() {
	_u16	index;
	char tempText[100];	
/*
	sprintf(tempText,"--------------------------------------------------------------------------");PrintToEmuOutput(tempText);
	sprintf(tempText,"ELF INFO");PrintToEmuOutput(tempText);
	sprintf(tempText,"Entry Point Adresse : %lx", ico_elf_hdr.entry_point_add );PrintToEmuOutput(tempText);
	sprintf(tempText,"\n");PrintToEmuOutput(tempText);	
	sprintf(tempText,"Start Of Program Header : %lx", ico_elf_hdr.start_prg_hdr );PrintToEmuOutput(tempText);
	sprintf(tempText,"Number Of  Program Header : %x", ico_elf_hdr.num_of_prg_hdr );PrintToEmuOutput(tempText);
	sprintf(tempText,"Size Of  Program Header : %x", ico_elf_hdr.prg_hdr_entry_size );PrintToEmuOutput(tempText);
*/
/*
	for( index = 0 ; index < ico_elf_hdr.num_of_prg_hdr; index++) {
		
		if ( ico_elf_prg_hdr[index].segment_type == 0x1 ) {

		sprintf(tempText,"ProgramHeader %d",index);PrintToEmuOutput(tempText);
		sprintf(tempText," Offset from file start to program segment : %lx", ico_elf_prg_hdr[index].offs_from_start_to_prg_seg );PrintToEmuOutput(tempText);
		sprintf(tempText," Virtual Adress of Segement : %lx", ico_elf_prg_hdr[index].virtual_add_of_segment );PrintToEmuOutput(tempText);
		sprintf(tempText," Physical Adress of Segment  : %lx", ico_elf_prg_hdr[index].physic_add_of_segment );PrintToEmuOutput(tempText);
		sprintf(tempText," Number of bytes in file image  : %lx ( %ld ) ", ico_elf_prg_hdr[index].num_of_bytes_in_file_ima, ico_elf_prg_hdr[index].num_of_bytes_in_file_ima );PrintToEmuOutput(tempText);
		sprintf(tempText," Number of bytes in mem image : %lx ( %ld ) ", ico_elf_prg_hdr[index].num_of_bytes_in_mem_ima, ico_elf_prg_hdr[index].num_of_bytes_in_mem_ima );PrintToEmuOutput(tempText);

		}
	}
*/
/*
printf(tempText,"\n");PrintToEmuOutput(tempText);
	sprintf(tempText,"Start Of Section Header : %lx ( %ld )", ico_elf_hdr.start_sct_hdr, ico_elf_hdr.start_sct_hdr  );PrintToEmuOutput(tempText);
	sprintf(tempText,"Number of  Section Header : %x ( %d )", ico_elf_hdr.num_of_sec_hdr ,  ico_elf_hdr.num_of_sec_hdr );PrintToEmuOutput(tempText);
	sprintf(tempText,"Section Header Entry Size: %x ( %d )", ico_elf_hdr.sec_hdr_entry_size,  ico_elf_hdr.sec_hdr_entry_size);PrintToEmuOutput(tempText);
	sprintf(tempText,"Section Header Stringatable Index: %x ( %d )\n", ico_elf_hdr.sec_hdr_stringa_index, ico_elf_hdr.sec_hdr_stringa_index );PrintToEmuOutput(tempText);
*/	
/*	
	for( index = 0 ; index < ico_elf_hdr.num_of_sec_hdr; index++) {
		ElfPrintSectionInfo( index );
		
	}	
*/		
//	sprintf(tempText,"--------------------------------------------------------------------------");PrintToEmuOutput(tempText);
}



void ElfPrintSectionInfo( _u8 fIndex ) {
	char tempText[100];	
/*
	sprintf(tempText,"--------------------------------------------------------------------------");PrintToEmuOutput(tempText);

	sprintf(tempText,"Section numero : %d",fIndex);PrintToEmuOutput(tempText);
	sprintf(tempText," Type : %lx ( %ld )", ico_elf_sec_hdr[fIndex].type, ico_elf_sec_hdr[fIndex].type );PrintToEmuOutput(tempText);
	sprintf(tempText," Flag : %lx ( %ld )", ico_elf_sec_hdr[fIndex].sect_flag, ico_elf_sec_hdr[fIndex].sect_flag );PrintToEmuOutput(tempText);
	sprintf(tempText," Numero Index Section : %lx ( %ld )", ico_elf_sec_hdr[fIndex].numero_index_sect,  ico_elf_sec_hdr[fIndex].numero_index_sect );PrintToEmuOutput(tempText);
	sprintf(tempText," Section Start Adress : %lx ( %ld )", ico_elf_sec_hdr[fIndex].sect_start_add, ico_elf_sec_hdr[fIndex].sect_start_add );PrintToEmuOutput(tempText);
	sprintf(tempText," Offset from start : %lx ( %ld )", ico_elf_sec_hdr[fIndex].offset_from_start ,  ico_elf_sec_hdr[fIndex].offset_from_start  );PrintToEmuOutput(tempText);
	sprintf(tempText," Size : %lx ( %ld ) ", ico_elf_sec_hdr[fIndex].size,  ico_elf_sec_hdr[fIndex].size  );PrintToEmuOutput(tempText);
	sprintf(tempText," Alignement : %lx ( %ld ) ", ico_elf_sec_hdr[fIndex].alignement,  ico_elf_sec_hdr[fIndex].alignement  );PrintToEmuOutput(tempText);
	sprintf(tempText," Fixed Size Entry : %lx ( %ld ) ", ico_elf_sec_hdr[fIndex].fixed_size_entry, ico_elf_sec_hdr[fIndex].fixed_size_entry);PrintToEmuOutput(tempText);

sprintf(tempText,"--------------------------------------------------------------------------");PrintToEmuOutput(tempText);
*/
}

