/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>

#include <math.h>
#include <string.h>


#include "nSX2.h"
//#include "../PS2Edefs.h"

#include "VU.h"
#include "VU1_micromode_tables.h"

#include "VIF.h"

  
/********************************************************************************************/
/*							VIF TRANSFER FUNCTION											*/
/********************************************************************************************/



#define	JumpToNextCode(A)	TransferSize = TransferSize - (A) - 1;CurrentData = CurrentData + (A) + 1



void VIF_Transfer( _u8	fVIF, _u32 * fAddr , _u32 fsize )  {

	_u32	VIFCode;
	_u32	VIFCode_Cmd;
	_u32	VIFCode_Num;
	_u32	VIFCode_Imm;
	_u32	VIFCode_iBit;

	_u32	TransferSize;
	_u32	* CurrentData;

	_u32	UnpackDataSize;
 

	/* Show debug Info */
#ifdef VIF_LOG_TR
	VIF_LOG_TR("[VIF%1.1d][TR]START size:%d (_u32) ", fVIF, fsize);
#endif

	/* Transfer Loop Initialisation */
	CurrentData = fAddr;
	TransferSize = fsize ; //Size unit is 32 bits 

 
	/* While not all the 32bit  DATA isn't transfered */
	while ( (_u32)TransferSize > (_u32)0 ) {

	/* Extraction of the Current VifCode */
	VIFCode 		= *(CurrentData);
	VIFCode_iBit 	= ( VIFCode >> 31 ) & 0x1;
	VIFCode_Cmd 	= ( VIFCode >> 24 ) & 0x7F;
	VIFCode_Num 	= ( VIFCode >> 16 ) & 0xFF;	
	VIFCode_Imm 	= ( VIFCode       ) & 0xFFFF;	

	if ( VIFCode_iBit == 1 ) {
#ifdef VIF_LOG_TR
		VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]iBit Detected", fVIF,TransferSize );
#endif
	}


	switch ( VIFCode_Cmd ) {

		/* No OPeration */
		case VIF_CMD_NOP: 		
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_NOP", fVIF,TransferSize );
#endif
			JumpToNextCode(0);
		break;

 
		/* Set Write Cycle*/
		case VIF_CMD_STCYCL: 	
			VIF[ fVIF ].Reg->cycle = (_u32)VIFCode_Imm;

#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_STCYCL CYCLE_REG:0x%8.8x", 
				fVIF,TransferSize,VIF[ fVIF ].Reg->cycle );
#endif

			JumpToNextCode(0);
		break;


		/* Set the double buffer Offset */
		case VIF_CMD_OFFSET: 	
			/* Write Imm to Offset Reg */
			VIF[ fVIF ].Reg->ofst = (_u32)(VIFCode_Imm & 0x3FF);

#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_OFFSET OFST_REG:0x%4.4lx", 
				fVIF,TransferSize,VIF[ fVIF ].Reg->ofst );
#endif

			/* Clear the DBF bit of VIF1_STAT */
			VIF[ fVIF ].Reg->stat = (_u32)( VIF[ fVIF ].Reg->stat & ~(0x80) );

			/* Copy BASE reg to TOPS reg */
			VIF[ fVIF ].Reg->tops = (_u32)VIF[ fVIF ].Reg->base;
			
			JumpToNextCode(0);

		break;


		/* Set the Base Adress of the Double Buffer */
		case VIF_CMD_BASE: 
			/* Write Imm to Base Reg */
			VIF[ fVIF ].Reg->base = (_u32)(VIFCode_Imm & 0x3FF);

#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_BASE BASE_REG:0x%4.4lx", 
				fVIF,TransferSize,VIF[ fVIF ].Reg->base );
#endif

			JumpToNextCode(0);

		break;


		case VIF_CMD_ITOP: 
			/* Write Imm to itop Reg */
			VIF[ fVIF ].Reg->itop = (_u32)(VIFCode_Imm & 0x3FF);

#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_ITOP ITOP_REG:0x%4.4lx", 
				fVIF,TransferSize,VIF[ fVIF ].Reg->itop );
#endif

			JumpToNextCode(0);
		break;


		case VIF_CMD_STMOD: 
			VIF[ fVIF ].Reg->mode = (_u32)(VIFCode_Imm & 0x3);

#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_STMOD MODE_REG:0x%4.4lx", 
				fVIF,TransferSize,VIF[ fVIF ].Reg->mode );
#endif

			JumpToNextCode(0);

		break;

/*
		case VIF_CMD_MSKPATH3: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_MSKPATH3", 
				fVIF,TransferSize);
#endif

			JumpToNextCode(0);
		break;
*/


		/* Muad check the docs - incomplete*/
		case VIF_CMD_MARK: 
			VIF[ fVIF ].Reg->mark = (_u32)( VIFCode_Imm & 0xFFFF );

#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_MARK MARK_REG:0x%4.4lx", 
				fVIF,TransferSize,VIF[ fVIF ].Reg->mark );
#endif

			JumpToNextCode(0);
		break;

		
		case VIF_CMD_FLUSHE: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_FLUSHE", fVIF,TransferSize);
#endif

			JumpToNextCode(0);
		break;


		case VIF_CMD_FLUSH: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_FLUSH", fVIF,TransferSize);
#endif

			JumpToNextCode(0);
		break;


		case VIF_CMD_FLUSHA	: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_FLUSHA", fVIF,TransferSize);
#endif

			JumpToNextCode(0);
		break;


		/* Activate MicroProgam */
		case VIF_CMD_MSCALF:
		case VIF_CMD_MSCAL: 

#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_MSCAL/F Add:0x%4.4lx", 
				fVIF,TransferSize,(VIFCode_Imm*8));
#endif

			/* Transfert TOPS/ITOPS to TOP/ITOP */
			VIF_InitMicroProgramActivation ( fVIF );

			switch ( fVIF ) {
				case VIF0: 
					/* Set the VectorUnit PC */	
					VectorUnit0.PC = (_u32)( VIFCode_Imm * 8 );
					/* The VectorUnit is running */
					VectorUnit0.halt = 0;
					/*Execute the MicroProgram*/
					VU0_Execute();	
				break;

				case VIF1:
					/* Set the VectorUnit PC */	
					VectorUnit1.PC = (_u32)( VIFCode_Imm * 8 );
					/* The VectorUnit is running */
					VectorUnit1.halt = 0;
					/*Execute the MicroProgram*/
					VU1_Execute();
				break;
			}

			JumpToNextCode(0);
		break;


		/* Continue Execution of the MicroProgram*/
		case VIF_CMD_MSCNT: 
#ifdef VIF_LOG_TR
		VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_MSCNT PC:0x%4.4lx", 
			fVIF,TransferSize,VectorUnit1.PC);
#endif
		
			/* Transfert TOPS/ITOPS to TOP/ITOP */
			VIF_InitMicroProgramActivation ( fVIF );

			switch ( fVIF ) {
				case VIF0: 
					/* The VectorUnit is running */
					VectorUnit0.halt = 0;
					/*Execute the MicroProgram*/
					VU0_Execute();		
				break;

				case VIF1:
					/* The VectorUnit is running */
					VectorUnit1.halt = 0;
					/*Execute the MicroProgram*/
					VU1_Execute();
				break;
			}


			JumpToNextCode(0);
		break;


		case VIF_CMD_STMASK	: 
			VIF[ fVIF ].Reg->mask = (_u32)( *(CurrentData + 1) );

#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_STMASK MASK_REG:0x%4.4lx", 
				fVIF,TransferSize,VIF[ fVIF ].Reg->mask );
#endif

			JumpToNextCode(1);

		break;


		case VIF_CMD_STROW: 

			VIF[ fVIF ].Reg->r0 = (_u32)( *(CurrentData + 1) );
			VIF[ fVIF ].Reg->r1 = (_u32)( *(CurrentData + 2) );
			VIF[ fVIF ].Reg->r2 = (_u32)( *(CurrentData + 3) );
			VIF[ fVIF ].Reg->r3 = (_u32)( *(CurrentData + 4) );

#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_STROW r0:0x%8.8lx r1:0x%8.8lx r2:0x%8.8lx r3:0x%8.8lx", 
				fVIF,TransferSize,VIF[ fVIF ].Reg->r0, VIF[ fVIF ].Reg->r1, 
				VIF[ fVIF ].Reg->r2, VIF[ fVIF ].Reg->r3 );
#endif

			JumpToNextCode(4);		
		break;


		case VIF_CMD_STCOL: 

			VIF[ fVIF ].Reg->c0 = (_u32)( *(CurrentData + 1) );
			VIF[ fVIF ].Reg->c1 = (_u32)( *(CurrentData + 2) );
			VIF[ fVIF ].Reg->c2 = (_u32)( *(CurrentData + 3) );
			VIF[ fVIF ].Reg->c3 = (_u32)( *(CurrentData + 4) );


#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_STCOL c0:0x%8.8lx c1:0x%8.8lx c2:0x%8.8lx c3:0x%8.8lx", 
				fVIF,TransferSize,VIF[ fVIF ].Reg->c0, VIF[ fVIF ].Reg->c1, 
				VIF[ fVIF ].Reg->c2, VIF[ fVIF ].Reg->c3 );
#endif

			JumpToNextCode(4);		
		break;


		/* Transfer the MicroProgram */
		case VIF_CMD_MPG: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_MPG Size: %d Addr:0x%4.4lx", 
				fVIF,TransferSize, VIFCode_Num, (VIFCode_Imm*8) );
#endif
			/* if size is O then size is 256 */
			if ( VIFCode_Num == 0 ) VIFCode_Num = 256 /*/ 2*/;

			/* Copy MicroProgram to VectorUnit */
			VIF_CopyMicroProgram ( fVIF, CurrentData , VIFCode_Num ) ;
 			
			/* Compute the next Data */
			JumpToNextCode(( VIFCode_Num * 2 ));
		break;


		case VIF_CMD_DIRECTHL: 
		case VIF_CMD_DIRECT	: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_DIRECT/HL Size: %d ", 
				fVIF,TransferSize, VIFCode_Num );
#endif

			/* if size is O then size is 65536 */
			if ( VIFCode_Imm == 0 ) VIFCode_Imm = 65536 / 4;

			/* PATH 2 */
			GIF_PATH2_Transfert((CurrentData+1),( VIFCode_Imm ));

			/* Compute the next Data */
			JumpToNextCode(( VIFCode_Imm * 4 ));
		break;

		

		/*********************************************/
		/*				UNPACK						*/
		/*********************************************/
/*
		case VIF_CMD_UNPACK_S32: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_S32 ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*	
		case VIF_CMD_UNPACK_S32_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_S32_M ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_S16: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_S16 ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_S16_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_S16_M ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_S8: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_S8 ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_S8_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_S8_M ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*

		case VIF_CMD_UNPACK_V2_32: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V2_32 ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V2_32_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V2_32_M ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V2_16: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V2_16", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V2_16_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V2_16_M ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V2_8: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V2_8 ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V2_8_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V2_8_M ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V3_32: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V3_32 ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V3_32_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V3_32_M ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*	
		case VIF_CMD_UNPACK_V3_16: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V3_16", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V3_16_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V3_16_M ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V3_8: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V3_8 ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V3_8_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V3_8_M ", fVIF,TransferSize );
#endif

			JumpToNextCode(?);
		break;
*/

		case VIF_CMD_UNPACK_V4_32: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V4_32 ", fVIF,TransferSize );
#endif
			/* Compute the Data Lenght */
			UnpackDataSize = VIF_ComputePackLenght( fVIF, 0x3 , 0x0  , VIFCode_Num);

			/* Unpak Data */
			VIF_Unpack_V4_32( fVIF, CurrentData , UnpackDataSize);
		
			/* Compute next VIF command */
			JumpToNextCode(UnpackDataSize);

		break;
/*
		case VIF_CMD_UNPACK_V4_32_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V4_32_M ", fVIF,TransferSize );
#endif
			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V4_16: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V4_16 ", fVIF,TransferSize );
#endif
			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V4_16_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V4_16_M ", fVIF,TransferSize );
#endif
			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V4_8: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V4_8 ", fVIF,TransferSize );
#endif
			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V4_8_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V4_8_M ", fVIF,TransferSize );
#endif
			JumpToNextCode(?);
		break;
*/
/*

		case VIF_CMD_UNPACK_V4_5: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V4_5 ", fVIF,TransferSize );
#endif
			JumpToNextCode(?);
		break;
*/
/*
		case VIF_CMD_UNPACK_V4_5_M: 
#ifdef VIF_LOG_TR
			VIF_LOG_TR("[VIF%1.1d][TR][%4.4d]CMD_UNPACK_V4_5_M ", fVIF,TransferSize );
#endif
			JumpToNextCode(?);
		break;
*/

		default:
#ifdef EMU_LOG
		EMU_LOG("[VIF%1.1d][TR][%4.4d]CMD_UNKNOW 0x%2.2x ", fVIF,TransferSize,VIFCode_Cmd );
#endif

		if ( VIF[ fVIF ].IgnoreVIFCodeError == TRUE ) {

			JumpToNextCode(0);

		} else {

			TransferSize = 0;

			switch ( fVIF ) {
				case VIF0: 
					INTC_RaiseIntByCause ( INTC_VIF0 );
				break;

				case VIF1:
					INTC_RaiseIntByCause ( INTC_VIF1 );
				break;
			}

		}
		break;


		}
	
	}


}

  












