/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include <strings.h> 
#include <float.h>
#include <math.h>

#include "nSX2.h"
#include "PS2Edefs.h" 

#include "r5900_main.h"
#include "r5900_core.h"	//for NI

#include "VU_cop2.h"
#include "VU_cop2_table.h"

#include "r5900_debug_regs.h"

#include "VU.h"

#include "VU_opcodes.h"
#include "VU1_micromode_debug.h"



/* LOG Function for the Interrupt Controller*/
void VU_COP2_Log ( char * ftext  , ...   ) {
	char VU_LogText[ FILE_NAME_LENGHT ];
	if ( debug_op == 1 ) {

	    va_list ap;
		va_start (ap, ftext);
	  	vsprintf (debugTempText, ftext, ap);
	   	va_end (ap);

		sprintf(debugTempText,"%s\n",debugTempText);
		

	}

}


 
/*****************************************************************************************************************
									COP2 INSTRUCTION SET
*****************************************************************************************************************/





void COP2()			{	cop2_opc[((_OPCODE_)&0x3E00000)>>21]();	}


/*
LQC2 : Floating-Point Data Transfer from EE Core to VU
*/
void    LQC2()		{	
	_u64 off;
	_u128	* temp128p;
	
	off = Extractoffset();
	sign_extendedh(off);

	temp128p = ( _u128 * ) &VectorUnit0.CPR[CURRENT_FT][0] ;
	MipsMemoryRead ( temp128p , DATA_QUAD ,  MipsCpu.GPR[CURRENT_BASE].lo , off ) ;

	INCREMENT_PC();

}

/*
QMTC2 : Floating-Point Data Transfert from EE to VU
*/
void cop2_QMTC2( ) {
	_u128	* temp128p;

	temp128p = ( _u128	*  ) &VectorUnit0.CPR[CURRENT_FS][0] ;
	temp128p->lo = (_u64)MipsCpu.GPR[CURRENT_RT].lo;
	temp128p->hi = (_u64)MipsCpu.GPR[CURRENT_RT].hi;

	INCREMENT_PC();
}


 


/*
SQC2 : Floating-Point Data Transfer from VU to EE Core
*/
void    SQC2()		{	
	_u64 off;
	_u128	* temp128p;

	off = Extractoffset();
	sign_extendedh(off);

	temp128p = ( _u128 * ) &VectorUnit0.CPR[CURRENT_FT][0] ;
	MipsMemoryWrite ( *temp128p , DATA_QUAD ,  MipsCpu.GPR[CURRENT_BASE].lo , off ) ;


	INCREMENT_PC();
}



/*
QMFC2 : Floating-Point Data Transfer from VU to EE Core
*/
void cop2_QMFC2() {
	_u128	* temp128p;

	temp128p = ( _u128	*  ) &VectorUnit0.CPR[CURRENT_FS][0] ;
	MipsCpu.GPR[CURRENT_RT].lo = (_u64)temp128p->lo;
	MipsCpu.GPR[CURRENT_RT].hi = (_u64)temp128p->hi;

	INCREMENT_PC();

}




/*
CFC2 : Transfer Integer Data From VU to EE Core
*/
void cop2_CFC2() {


	MipsCpu.GPR[CURRENT_RT].lo =  (_u32)VectorUnit0.CCR[ CURRENT_ID ] ;

	INCREMENT_PC();
}

  

/*
CTC2 : Transfer Integer Data from EE Core to VU
*/
void cop2_CTC2	() {


	VectorUnit0.CCR[ CURRENT_ID ] = (_u32)MipsCpu.GPR[CURRENT_RT].lo ;

	INCREMENT_PC();
}

    
/*	
BC2 : Branch on COP2 Conditional Signal
*/
//void cop2_BC() { 	PrintDebugOpcode() ; MipsCpu.halt = 1;}
void cop2_BC() { 	cop2_BC2_opc[( _OPCODE_ >>16)& 0x3 ](); } 
void BC2F() { 	EMU_LOG ( "BC2F" ); PrintDebugOpcode() ; MipsCpu.halt = 1;}
void BC2T() { 	EMU_LOG ( "BC2T" ); PrintDebugOpcode() ; MipsCpu.halt = 1;}
void BC2FL() { 	EMU_LOG ( "BC2FL" ); PrintDebugOpcode() ; MipsCpu.halt = 1;}
void BC2TL() { 	EMU_LOG ( "BC2TL" ); PrintDebugOpcode() ; MipsCpu.halt = 1;}


/*****************************************************************************************************************
									COP0 VU0 INSTRUCTION SET
*****************************************************************************************************************/

void cop2_Vopc() 	{ 
	_VU_OPCODE_ = (_u32)_OPCODE_;
	cop2_V_opc[((_VU_OPCODE_)&0x3F)](); 

} 
 
void Vopc_t60()	{ 	cop2_V_t60_opc[((_VU_OPCODE_)&0x7C0)>>6]();	} 
void Vopc_t61()	{ 	cop2_V_t61_opc[((_VU_OPCODE_)&0x7C0)>>6]();	} 
void Vopc_t62()	{ 	cop2_V_t62_opc[((_VU_OPCODE_)&0x7C0)>>6]();	} 
void Vopc_t63()	{ 	cop2_V_t63_opc[((_VU_OPCODE_)&0x7C0)>>6](); } 

            
void 	VABS()		{ 	
	vuABS(VectorUnit0);
	INCREMENT_PC();
}

void 	VADD()		{ 
	vuADD(VectorUnit0);
	INCREMENT_PC();
} 

void 	VADDi()		{ 	
	vuADDi(VectorUnit0);
	INCREMENT_PC();
} 

void	VADDq()		{ 
	vuADDq(VectorUnit0);
	INCREMENT_PC();
} 
  
void 	VADDbc()	{
	vuADDbc(VectorUnit0);
	INCREMENT_PC();
} 

void 	VADDA()		{ 
	vuADDA(VectorUnit0);
	INCREMENT_PC();
}  
  
void 	VADDAi()	{ 	
	vuADDAi(VectorUnit0);
	INCREMENT_PC();
} 

void 	VADDAq()	{ 	
	vuADDAq(VectorUnit0);
	INCREMENT_PC();
} 

void 	VADDAbc()	{ 	
	vuADDAbc(VectorUnit0);
	INCREMENT_PC();
} 

void 	VCALLMS()	{ 	PrintDebugOpcode() ; MipsCpu.halt = 1; } 
void 	VCALLMSR()	{ 	PrintDebugOpcode() ; MipsCpu.halt = 1; } 


void 	VCLIP()		{ 	
	vuCLIP(VectorUnit0)	
	INCREMENT_PC();
}


void	VDIV()		{
	vuDIV(VectorUnit0);
	INCREMENT_PC();
}  

void 	VFTOI0()	{ 
	vuFTOI0(VectorUnit0);
	INCREMENT_PC();	
}  

void 	VFTOI4()	{ 	
	vuFTOI4(VectorUnit0);
	INCREMENT_PC();	
}

void 	VFTOI15()	{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;	}
void 	VFTOI12()	{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;	} 
 
void 	VIADD()		{ 	
	vuIADD(VectorUnit0);
	INCREMENT_PC();	
} 

void 	VIADDI()	{ 	
	vuIADDI(VectorUnit0)
	INCREMENT_PC();	
} 

void 	VIAND()		{ 
	vuIAND(VectorUnit0);
	INCREMENT_PC();
}  

void 	VILWR()		{ 	
	vuILWR(VectorUnit0);
	INCREMENT_PC();
}

  
void 	VIOR()		{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;	}
void 	VISUB()		{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;}

 
void 	VISWR()		{ 		
	vuISWR(VectorUnit0);
	INCREMENT_PC();	
} 
 
void 	VITOF0()	{ 	
	vuITOF0(VectorUnit0);
	INCREMENT_PC();	
} 


void 	VITOF4()	{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;	}
void 	VITOF12()	{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;	} 
void 	VITOF15()	{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;	}
void 	VLQD()		{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;	}  


void 	VLQI()		{ 		
	vuLQI(VectorUnit0);
	INCREMENT_PC();
}
   
void 	VMADD()		{ 	
	vuMADD(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMADDi()	{ 	
	vuMADDi(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMADDq()	{ 	
	vuMADDq(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMADDbc()	{ 
	vuMADDbc(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMADDA()	{ 	
	vuMADDA(VectorUnit0);
	INCREMENT_PC();
}

void 	VMADDAi()	{ 	
	vuMADDAi(VectorUnit0);
	INCREMENT_PC();
}

void 	VMADDAq()	{ 
	vuMADDAq(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMADDAbc()	{  
	vuMADDAbc(VectorUnit0)
	INCREMENT_PC();
}

void 	VMAX()		{ 	
	vuMAX(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMAXi()		{ 	
	vuMAXi(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMAXbc()	{ 	
	vuMAXbc(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMFIR()		{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;
}
void 	VMINI()		{ 	
	vuMINI(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMINIi()	{ 
	vuMINIi(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMINIbc()	{ 	
	vuMINIbc(VectorUnit0);	
	INCREMENT_PC();
} 

void 	VMOVE()		{ 
	vuMOVE(VectorUnit0);
	INCREMENT_PC();
}  

void 	VMR32()		{ 	
	vuMR32(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMSUB()		{ 		
	vuMSUB(VectorUnit0);
	INCREMENT_PC();
}
 
void 	VMSUBi()	{ 	
	vuMSUBi(VectorUnit0);
	INCREMENT_PC();
}

void 	VMSUBq()	{ 
	vuMSUBq(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMSUBbc()	{	
	vuMSUBbc(VectorUnit0);
	INCREMENT_PC();
}  

void 	VMSUBA()	{ 	
	vuMSUBA(VectorUnit0);
	INCREMENT_PC();
}

void 	VMSUBAi()	{ 		
	vuMSUBAi(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMSUBAq()	{ 		
	vuMSUBAq(VectorUnit0);
	INCREMENT_PC();
}
 
void 	VMSUBAbc()	{ 	
	vuMSUBAbc(VectorUnit0);
	INCREMENT_PC();
}
 
void 	VMTIR()		{ 	PrintDebugOpcode() ; MipsCpu.halt = 1;	}  

void 	VMUL()		{
	vuMUL(VectorUnit0);
	INCREMENT_PC();
}

void 	VMULi()		{ 	
	vuMULi(VectorUnit0);
	INCREMENT_PC();
}

void 	VMULq()		{ 
	vuMULq(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMULbc()	{
	vuMULbc(VectorUnit0);
	INCREMENT_PC();
} 
 
void 	VMULA()		{		
	vuMULA(VectorUnit0);
	INCREMENT_PC();
}

void 	VMULAi()	{ 		
	vuMULAi(VectorUnit0);
	INCREMENT_PC();
}
  
void 	VMULAq()	{ 	
	vuMULAq(VectorUnit0);
	INCREMENT_PC();
} 

void 	VMULAbc()	{ 
	vuMULAbc(VectorUnit0);
	INCREMENT_PC();
}

void 	VNOP()		{ 	
	vuNOP(VectorUnit0);	
	INCREMENT_PC();	
} 

void 	VOPMULA()	{ 	
	vuOPMULA(VectorUnit0);
	INCREMENT_PC();
}  

void 	VOPMSUB()	{ 	
	vuOPMSUB(VectorUnit0);
	INCREMENT_PC();
} 
 
void 	VRGET()				// fix for pills demo
{
	vuRGET(VectorUnit0);
	INCREMENT_PC();
} 
	
void 	VRINIT()	{
	vuRINIT(VectorUnit0)
	INCREMENT_PC();
}  
		
void 	VRNEXT()	
{
	vuRNEXT(VectorUnit0)	// fix for pills demo
	INCREMENT_PC();
}  

void 	VRSQRT()	{
	vuRSQRT(VectorUnit0);
	INCREMENT_PC();
}  

void 	VRXOR()				// fix for pills demo
{
	vuRXOR(VectorUnit0);
	INCREMENT_PC();
} 

void 	VSQD()		{ 		
	vuSQD(VectorUnit0);
	INCREMENT_PC();
}

void 	VSQI()		{ 
	vuSQI(VectorUnit0);
	INCREMENT_PC();
}

void 	VSQRT()		{ 	
	vuSQRT(VectorUnit0);
	INCREMENT_PC();
}

void 	VSUB()		{ 	
	vuSUB(VectorUnit0);
	INCREMENT_PC();
} 

void 	VSUBi()		{ 		
	vuSUBi(VectorUnit0);
	INCREMENT_PC();
}
 
void 	VSUBq()		{ 	
	vuSUBq(VectorUnit0);
	INCREMENT_PC();
} 

void 	VSUBbc()	{ 
	vuSUBbc(VectorUnit0);
	INCREMENT_PC();
} 

void 	VSUBA()		{ 		
	vuSUBA(VectorUnit0);
	INCREMENT_PC();
}
  
void 	VSUBAi()	{ 	
	vuSUBAi(VectorUnit0);
	INCREMENT_PC();
}
  
void 	VSUBAq()	{ 	
	vuSUBAq(VectorUnit0);
	INCREMENT_PC();
}
  
void 	VSUBAbc()	{ 	
	vuSUBAq(VectorUnit0);
	INCREMENT_PC();
}    

void 	VWAITQ()	{ 	
	vuWAITQ(VectorUnit0);	
	INCREMENT_PC();	
} 



	









