/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <fcntl.h>
#ifdef __WIN32__
#include <io.h>
#else
#include <unistd.h>
#endif
#include <string.h>

#include "../PsxCommon.h"
#include "Rpc_services.h"
#include "../CDVDisodrv.h"

#define IOPMEMSIZE	(2*1024*1024)

#define EIGHTMEGSm256	(8*1024*1024-256)	//0x007FFF00
#define ALIGN256		0xFFFFFF00

#define mALLOCATED(info)	 (info & 1)
#define mADDRESS(info)		((info >> 1) & 0x7FFF)
#define mSIZE(info)		 (info >> 17)
//[000000000000000 0  000000000000000 1]
//[ size of block |?|address of block|a]located
//upper 15bits are the size of block in 256bytes units [8MB addressability]
//unknown 1 bit
//15bits for address of block in 256bytes units [8MB addressability; 15+8 bits]
//the lower bit means that block is allocated or not (1/0)

struct allocELEM{	//sizeof()=8
	struct allocELEM	*next;
/*	unsigned int    size     :15,
			unknown  :1,	// info could be declared this way
			addr     :15,
			allocated:1;	*/
	unsigned int		info;	// here is stored data about
					// an allocable memory chunk
};

#define smFIRST		0
#define smFIRSTF	2		//first one that can be freed
#define smCHECK		27
#define smLAST		30
#define smMAX		(smLAST+1)	//max number of elements = 31

struct allocTABLE{	//sizeof()=256=0x100
	struct allocTABLE	*next;		//+00
	struct allocELEM	list[31];	//+04 8 * 31 = 248 = 0xF8;
	int					dummy;
};

unsigned int		memsize;
struct allocTABLE	*alloclist=NULL;

///////////////////////////////////////////////////////////////////////
void *sysmem_init_memory(){	//mark all mem as not allocated, available
	if (alloclist){
		unsigned int		i;
		struct allocELEM	*p;

		alloclist->next=NULL;
		for (i=smFIRST; i<smMAX; i++){
			alloclist->list[i].next=&alloclist->list[i+1];
			alloclist->list[i].info=0;
		}
		alloclist->list[smLAST].next=NULL;

		alloclist->list[smFIRST].info &= 0x0001FFFF;
		alloclist->list[smFIRST].info |= (((memsize<0 ?
					 memsize+255:memsize) >> 8) << 17);

		if (AllocSysMemory(ALLOC_FIRST, (unsigned int)alloclist-(unsigned int)PSXM(0), NULL)==0)	//this is not NULL! but the starting address of the first 0x1500 bytes allocated
			if (((int)alloclist-(int)PSXM(0))==(int)AllocSysMemory(ALLOC_FIRST, sizeof(struct allocTABLE)-4, NULL)){ //alloc allocated allocation table;-)
				for (p=alloclist->list; p; p=p->next)
					if (!mALLOCATED(p->info))
						return (void*)(PSXM((mADDRESS(p->info))<<8));//next free block address
				return NULL;
			}
		alloclist=NULL;
	}
	return NULL;
}

///////////////////////////////////////////////////////////////////////
void *sysmemInit(unsigned int iopmemsize){
	iopmemsize=min(EIGHTMEGSm256, iopmemsize);

	alloclist = (struct allocTABLE*)PSXM(0x1500);

	memsize = iopmemsize & 0xFFFFFF00;

	if (memsize >= ((unsigned int)alloclist-(unsigned int)PSXM(0)
					+sizeof(struct allocTABLE)))	//alloctable must fit memory
		return sysmem_init_memory();
	
	return alloclist = NULL;
}

///////////////////////////////////////////////////////////////////////
void *allocmem(int flags, int size, void *mem){
     struct allocELEM	*a;//, *last, *k;
     unsigned int	bsize,		//block size in 256bytes units
//					baddress,	//block address in 256bytes units
					i, tmp;
     void		*address;

	bsize = (size+255) >> 8;	// round to upper
	if (bsize==0)		return NULL;

	switch (flags){
	case ALLOC_FIRST:
		for (a = alloclist->list; a; a=a->next)
			if ((!mALLOCATED(a->info)))
				if (mSIZE(a->info) >= bsize)
					break;
		if (a==NULL) return NULL;

		if (mSIZE(a->info) == bsize){
			a->info|=1;			//alloc it
			return (void*)(mADDRESS(a->info) << 8);	//all done, how quick!
		}

		address = (void*)(mADDRESS(a->info) << 8);
		i = a->info;

		a->info=((a->info | 1) & 0x1FFFF) |	//alloc block
			  (bsize << 17);		//of wanted size

		i = (i & 0xFFFF0001) | (((mADDRESS(i) + bsize) & 0x7FFF) << 1);
		i = (i & 0x1FFFF) | ((mSIZE(i) - bsize) << 17);

		a=a->next;
		while (mSIZE(i)>0){
			tmp=a->info;
			a->info=i;
			i=tmp;
			a = a->next;
		}
		return address;

/*	case ALLOC_LAST:
		last=NULL;
		for (a = alloclist->list; a; a=a->next)
			if ((!mALLOCATED(a->info)))
				if (mSIZE(a->info) >= bsize)
					last=a;		//use last one suitable
		a = last;
		if (a==0) return NULL;

		if (mSIZE(a->info) == bsize){
			a->info|=1;			//alloc it
			return (void*)(mADDRESS(a->info) << 8);
		}

		a->info = (a->info & 0x0001FFFF) | 
		   ((mSIZE(a->info) - bsize) << 17);	//put rest of block

		i =  	(((i & 0xFFFF0001)//this line has no use; this is stupid
			| ((mADDRESS(a->info) + mSIZE(a->info)) & 0x7FFF) << 1)	//calculate address of block to use
			& 0x0001FFFF)	//keep only address (and alloc' state)
			| (bsize << 17)	//put size
			| 1;		//mark as allocated
		a=a->next;

		address = (void*)(mADDRESS(i) << 8); //address in bytes
		while (mSIZE(i)>0){
			tmp=a->info;
			a->info=i;
			i=tmp;
			a = a->next;
		}
		return address;

	case ALLOC_LATER:
		if ((unsigned int)mem & 0xFF) return NULL;
		baddress = (unsigned int)mem >> 8;		//addr in 256bytes units
		for (a = alloclist->list; a ; a=a->next){
			if (baddress < mADDRESS(a->info)) return NULL;//cannot realloc
			if ((!mALLOCATED(a->info)) &&
			     (mSIZE(a->info) + mADDRESS(a->info) >= baddress + bsize))
				break;
		}
		if (a==0) return NULL;
                
		if (mADDRESS(a->info) < baddress){
			tmp = mADDRESS(a->info) + mSIZE(a->info) - baddress;
			a->info= (a->info & 0x1FFFF) |
			   ((mSIZE(a->info) - tmp) << 17);

			i =(((i & 0xFFFF0001) |	//stupid compiler
			     ((mADDRESS(a->info) + mSIZE(a->info)) & 0x7FFF) << 1)
			    & 0x1FFFE)
			    | (tmp << 17);

			k = a = a->next;
			while (mSIZE(i)>0){
				tmp=a->info;
				a->info=i;
				i=tmp;
				a=a->next;
			}
			a = k;
		}

		if (mSIZE(a->info)==bsize){
			a->info|=1;			//alloc it
			return (void*)(mADDRESS(a->info) << 8);
		}

		address = (void*)(mADDRESS(a->info) << 8);
		i = a->info;

		a->info=((a->info | 1) & 0x1FFFF) |	//alloc block
			  (bsize << 17);

		i = (i & 0xFFFF0001) | (((mADDRESS(i) + bsize) & 0x7FFF) << 1);
		i = (i & 0x1FFFF) | ((mSIZE(i) - bsize) << 17);

		a=a->next;
		while (mSIZE(i)>0){
			tmp=a->info;
			a->info=i;
			i=tmp;
			a = a->next;
		}
		return address;
*/
	}

	return NULL;
}

///////////////////////////////////////////////////////////////////////[OK]
int  freemem(void *mem){
     int		skip;	// counter (0,1,2) of elements to remove
     struct allocELEM	*a,	// an element
			*p,	// prev
			*n;	// next

	if ((unsigned int)mem & 0xFF)		return -1;	// only 256byte multiple

	for (a=&alloclist->list[smFIRSTF], p=NULL; a; a=a->next){
		if ((mSIZE(a->info) && (mADDRESS(a->info)==
						((unsigned int)mem >> 8))))	// convert to 256byte units
			break;
		p=a;
	}

	if (a==NULL)		return -1;	//block not found

	if (!mALLOCATED(a->info))return -1;	//cannot free a freed block

	n=NULL;	
	skip=0;	

	a->info&=0xFFFFFFFE;	//free block

	if ((a->next) && (!mALLOCATED(a->next->info))){//bind with next free block
		n = a->next;
		skip = 1;
		a->info = (a->info & 0x1FFFE) |
		    ((mSIZE(a->info) + mSIZE(a->next->info)) << 17);
	}

	if (p && (!mALLOCATED(p->info))){	//bind with previous free block
		n=a;
		skip++;			// or skip=2;
		p->info=(p->info & 0x1FFFF) | 
		  ((mSIZE(p->info) + mSIZE(a->info)) << 17);
	}

	if (skip){
		a=n;
		while (--skip!=-1)
			a=a->next;
		while (a){
			n->info=a->info;
			a=a->next;
			n=n->next;
		}
	}
	return 0;
}

///////////////////////////////////////////////////////////////////////[OK]
void maintain(){
     struct allocTABLE	*table, *new;
     int		i;

	for (table=alloclist; table->next; table = table->next)	//+0x00
		;	//move to last non-NULL
	if (mSIZE(table->list[smCHECK].info)>0)		//+0xE0	//27 (i.e.28th)
		// a new table is needed; alloc it
		if ((new=(struct allocTABLE*)allocmem(0, sizeof(struct allocTABLE), 0))){
			new=PSXM((int)new);
			table->next=new;		// link it
			table->list[smLAST].next=&new->list[smFIRST];// link it's table elements
			table=table->next;		// move to it
			table->next=NULL;		// mark as end of tables
			for (i=0; i<smMAX; i++){	// init table elements
				table->list[i].next=&table->list[i+1];
				table->list[i].info=0;
			}
			table->list[smLAST].next=NULL;
		}

	table=alloclist;
	if (table->next){
		while (table->next->next)
			table=table->next;
		if (table->next)	//redundant check
			if (mSIZE(table->list[smCHECK].info)==0){
				struct allocTABLE *t=table->next;
				table->list[smLAST].next=NULL;
				table->next		=NULL;
				freemem((void*)((int)t-(int)PSXM(0)));
			}
	}
}

///////////////////////////////////////////////////////////////////////[OK]
void *AllocSysMemory(int flags, int size, void *mem){

	if (alloclist==NULL)
		sysmemInit(IOPMEMSIZE);

	if (alloclist && (flags<3)){
		void *r=allocmem(flags, size, mem);
		maintain();
		return (void*)r;
	}
	return NULL;
}

///////////////////////////////////////////////////////////////////////[OK]
int  FreeSysMemory(void *mem){
     struct allocTABLE	*table;
     int		r;

 	if (alloclist==NULL)
		sysmemInit(IOPMEMSIZE);

	for (table=alloclist; table; table=table->next)	//must not be among
		if (mem==table)				//the allocation tables
			return -1;
	if ((r=freemem(mem)))
		return r;
	maintain();
	return r;	//r==0 ;-)
}

///////////////////////////////////////////////////
//IOPHEAP_ALLOC	1
//void *sceSifAllocIopHeap(int size);
void sysmem_alloc(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
#ifdef RPC_LOG
	RPC_LOG("[RPC:sysmem] AllocSysMemory(ALLOC_FIRST, size=%d, NULL)\n", *bin);
#endif
	*bout=(int)AllocSysMemory(ALLOC_FIRST, *bin, NULL);
#ifdef RPC_LOG
	RPC_LOG("             addr=0x%08X\n", *bout);
#endif
}

///////////////////////////////////////////////////
//IOPHEAP_FREE	2
//int sceSifFreeIopHeap(void *addr);
void sysmem_free(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
#ifdef RPC_LOG
	RPC_LOG("[RPC:sysmem] FreeSysMemory(addr=0x%08X)\n", *bin);
#endif
	*bout=FreeSysMemory((void*)*bin);
}

///////////////////////////////////////////////////
//IOPHEAP_LOAD	3
//int sceSifLoadIopHeap(char *fname, void *addr);
void sysmem_load(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
	register struct fileio_LOADIOPHEAP *in=(struct fileio_LOADIOPHEAP*)bin;
	register int fd;
	register int fsize;
	int		buffer=(int)in->buffer & 0xFFFFFF;
	char	*name, *device;

#ifdef RPC_LOG
	RPC_LOG("[RPC:sysmem] LoadIopHeap(\"%s\", addr=0x%08X)\n", in->name, in->buffer);
#endif

	name=device=in->name;while(*name && *name!=':')name++;if (*name==':'){*name=0; name++;}
	if ((strncmp(device, "rom", strlen("rom"))) &&
		(strncmp(device, "cdrom", strlen("cdrom"))) &&
		(strncmp(device, "host", strlen("host")))){
#ifdef RPC_LOG
		RPC_LOG("             unsuported device: \"%s\"\n", device);
#endif
		*bout=-1;
		return;
	}else if (strncmp(device, "cdrom", strlen("cdrom"))==0){
		if ((fd = CDVDFS_open(name, 1))<0){
#ifdef RPC_LOG
			RPC_LOG("             failed\n");
#endif
			*bout = -1;
		}

#ifdef RPC_LOG
		RPC_LOG("             %s => loaded\n", name);
#endif
		fsize = CDVDFS_lseek(fd, 0, SEEK_END);
		CDVDFS_lseek(fd, 0, SEEK_SET);
//		CDVDFS_read(fd, PSXM(buffer), fsize);    //read whole file
////////////////////////PSXM is buggy? this function is used by SvenGoranEricsonWorldManager
		CDVDFS_close(fd);
	}else{
		if ((*name!='\\') && 
			(*name!='/') && 
			(strncmp(name+1, ":\\", 2))){
#ifdef	__LINUX__
			strncpy(name-5, "host/", 5);		//dirty, very dirty;)
#else
			strncpy(name-5, "host\\", 5);
#endif
			name-=5;
		}
		
		if ((fd = open(name, O_RDONLY))<0){
#ifdef RPC_LOG
			RPC_LOG("             failed\n");
#endif
			*bout = -1;
		}

#ifdef RPC_LOG
		RPC_LOG("             %s => loaded\n", name);
#endif
		fsize = lseek(fd, 0, SEEK_END);
		lseek(fd, 0, SEEK_SET);
//		read(fd, PSXM(buffer), fsize);    //read whole file
/////////////////PSXM is buggy? this function is used by SvenGoranEricsonWorldManager
		close(fd);
	}

	*bout = 0;
}

///////////////////////////////////////////////////
//IOPHEAP_ALLOCSYSMEMORY	4
//void *sceSifAllocSysMemory(int mode, int size, void *addr);
void sysmem_allocsys(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
//	register struct ... *in=(struct ...*)bin;
#ifdef RPC_LOG
//	RPC_LOG("[RPC:sysmem] AllocSysMemory(mode=%d, size=%d, addr=0x%08X)\tNOT IMPLEMENTED\n", in->mode, in->size, in->mem);
#endif
	//*bout=(int)AllocSysMemory(in->mode, in->size, in->mem);
#ifdef RPC_LOG
	RPC_LOG("             addr=0x%08X\n", *bout);
#endif
}

//ENTRYPOINT
void sifcall_sysmem(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
	switch(call->rpc_number){
	case IOPHEAP_ALLOC:			sysmem_alloc(call, bin, bout);	break;
	case IOPHEAP_FREE:			sysmem_free(call, bin, bout);	break;
	case IOPHEAP_LOAD:			sysmem_load(call, bin, bout);	break;
	case IOPHEAP_ALLOCSYSMEMORY:sysmem_allocsys(call, bin, bout);break;
	default:
#ifdef RPC_LOG
		RPC_LOG("[RPC:sysmem] <%d> function is not implemented (0x%04X)\n", call->rpc_number, FILEIO_VER);
#endif
		*bout=-1;	//general fail
	}
}