/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include <strings.h>
#include <float.h>
#include <math.h>


#include "nSX2.h"

#include "r5900_main.h"
 
#include "bios_syscalls.h"
#include "bios_syscalls_table.h"
 
#include "PS2Edefs.h"

#include "Sif.h"

/* I must declare a function to avoid this kind of include */
#include "DmaC.h"

/* must create function to avoid this kind of declaration */
#include "IntrC.h"


#include "GS.h"




/****************************************************************************************************/





extern void (CALLBACK* SIFSetRPCVersion)(char* _ioprp); 

/*****************************************************************************************************************
									SYSCALL INSTRUCTION SET
*****************************************************************************************************************/

/* LOG Function for the Interrupt Controller*/
void SYSCALL_Log ( char * ftext  , ...   ) {
	char SYSCALL_LogText[ FILE_NAME_LENGHT ];

	if ( debug_systemcall == 1 ) {

    	va_list ap;
		va_start (ap, ftext);
  		vsprintf (SYSCALL_LogText, ftext, ap);
   		va_end (ap);

		sprintf(SYSCALL_LogText,"%s\n",SYSCALL_LogText);
		PrintToBIOSOutput 		(  SYSCALL_LogText ) ;
	}

}




/*
Show Syscall Name in the window
*/
void SYSCALL_PrintDebugInfo() {
	char temp_string[256];
	_u8		SystemCallNumber;	
 	
  	SystemCallNumber = MipsCpu.GPR[CPU_REG_V1].lo & 0xFF;
	
 	if ((signed char)SystemCallNumber < 0) SystemCallNumber =(_u8)((- SystemCallNumber ) & 0xFF);
 	
	sprintf(temp_string,"%8.8llx:%s\n", (long)MipsCpu.PC , debug_syscall[SystemCallNumber]);
	PrintToBIOSOutput (  temp_string );
}


/*
Syscall No Implemented
*/
void SYSCALL_NI( ) {

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]!!! SYSCALL UNKNOW !!! ",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1;
}

/*

*/
void SYSCALL_RFU000_FullReset() { 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU000_FullReset : NI ",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
} 




/*
ResetEE	ENTRY_1	8000ad68

*/
void SYSCALL_ResetEE( )	{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ResetEE : NI ",(long)MipsCpu.PC );
#endif	

//	MipsCpu.PC = 	MipsCpu.PC + 4;	

	MipsCpu.halt = 1; 
}  


/*


*/
void SYSCALL_RFU003( ) { 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU003 : NI ",(long)MipsCpu.PC );
#endif	

	
	MipsCpu.halt = 1; 
} 


/*

*/ 
void SYSCALL_Exit( ) 							{ 

	static	AlreadyPrint = 0;


//	if ( debug_print_syscall == 1 )	SYSCALL_PrintDebugInfo();
//	if ( !AlreadyPrint ) {
 		
#ifdef EMU_LOG
	EMU_LOG("Interpreter Stop ( Syscall Exit )");	
#endif	

// 		AlreadyPrint = 1;
//   		}
 		
	MipsCpu.halt = 1; 
	
} 


/*
RFU005	ENTRY_5	80002880
*/ 
void SYSCALL_RFU005( ) 							{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU005 : NI ",(long)MipsCpu.PC );
#endif	
	

	MipsCpu.PC = 	MipsCpu.PC + 4;	 
}


/*
LoadPS2Exe	ENTRY_6	800055a0

*/  
void SYSCALL_LoadPS2Exe( ) 						{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]LoadPS2Exe : NI ",(long)MipsCpu.PC );
#endif	

	

	MipsCpu.halt = 1; 
} 


/*
ExecPS2	ENTRY_7	80002f80


*/ 
void SYSCALL_ExecPS2( ) 						{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ExecPS2 : NI ",(long)MipsCpu.PC );
#endif	


MipsCpu.halt = 1; 
}

/*
RFU008	ENTRY_8

*/  
void SYSCALL_RFU008( ) 							{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU008 : NI ",(long)MipsCpu.PC );
#endif	
	

MipsCpu.halt = 1; 
}  

/*
RFU009	ENTRY_9	80005b28
*/
void SYSCALL_RFU009( ) 							{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU009 : NI ",(long)MipsCpu.PC );
#endif	

	

MipsCpu.halt = 1; 
}  

/*
AddSbusIntcHandler	ENTRY_10	80001e78

*/
void SYSCALL_AddSbusIntcHandler( ) 				{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]AddSbusIntcHandler : NI ",(long)MipsCpu.PC );
#endif	

	
MipsCpu.halt = 1; 
}

/*
RemoveSbusIntcHandler	ENTRY_11	80001ed0


*/  
void SYSCALL_RemoveSbusIntcHandler( ) 			{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RemoveSbusIntcHandler : NI ",(long)MipsCpu.PC );
#endif	

	
MipsCpu.halt = 1; 
}

/*
Interrupt2Iop	ENTRY_12	80001f70

*/  
void SYSCALL_Interrupt2Iop( ) 					{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]Interrupt2Iop : NI ",(long)MipsCpu.PC );
#endif	

	
MipsCpu.halt = 1; 
}


/*
SetVTLBRefillHandler	ENTRY_13	800009c0


*/  
void SYSCALL_SetVTLBRefillHandler( ) 			{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SetVTLBRefillHandler : NI ",(long)MipsCpu.PC );
#endif	


	
	INCREMENT_PC();
}

/*
SetVCommonHandler	ENTRY_14	80000a00


*/  
void SYSCALL_SetVCommonHandler( ) 				{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SetVCommonHandler : NI ",(long)MipsCpu.PC );
#endif	




	INCREMENT_PC();
}  

/*
SetVInterruptHandler	ENTRY_15	80000a40


*/
void SYSCALL_SetVInterruptHandler( ) 			{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SetVInterruptHandler : NI ",(long)MipsCpu.PC );
#endif	


	

MipsCpu.halt = 1; 
}  








/**********************************************************************************************************
																	
***********************************************************************************************************/





  



void SYSCALL_RFU063( ) 							{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU063 : NI",(long)MipsCpu.PC );
#endif	

	

MipsCpu.halt = 1; 
}  




void SYSCALL_KSeg0( ) 							{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]KSeg0 : NI",(long)MipsCpu.PC );
#endif	



	
	MipsCpu.halt = 1; 
}   



void SYSCALL_EnableCache( ) 					{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]EnableCache",(long)MipsCpu.PC );
#endif	


	//MipsCpu.halt = 1; 

	MipsCpu.PC = MipsCpu.PC + 4;
}  



void SYSCALL_DisableCache( ) 					{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]DisableCache",(long)MipsCpu.PC );
#endif	
	//MipsCpu.halt = 1; 

	MipsCpu.PC = MipsCpu.PC + 4;
}   



void SYSCALL_GetCop0( ) 						{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]GetCop0 : NI",(long)MipsCpu.PC );
#endif	



	MipsCpu.halt = 1; 
}   

/*
FlushCache 0x64 100
Do Nothing
*/
void SYSCALL_FlushCache( ) 						{ 
    static _u64 NbFlushCache = 0;;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]FlushCache",(long)MipsCpu.PC );
#endif	

	NbFlushCache++;
    MipsCpu.PC = MipsCpu.PC + 4;
}   



void SYSCALL_CpuConfig( ) 						{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]CpuConfig : NI",(long)MipsCpu.PC );
#endif	
	


	MipsCpu.halt = 1; 
}   






void SYSCALL_SetCPUTimerHandler( ) 				{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SetCPUTimerHandler : NI",(long)MipsCpu.PC );
#endif	
	
	

	MipsCpu.halt = 1; 
}   



void SYSCALL_SetCPUTimer( ) 					{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SetCPUTimer : NI",(long)MipsCpu.PC );
#endif	


	MipsCpu.halt = 1; 
}   



void SYSCALL_ForceRead( ) 						{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ForceRead : NI",(long)MipsCpu.PC );
#endif	


		MipsCpu.halt = 1; 
}   



void SYSCALL_ForceWrite( ) 						{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ForceWrite : NI",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
}   






void SYSCALL_SetPgifHandler( ) 					{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SetPgifHandler : NI",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
}    



void SYSCALL_SetVSyncFlag( ) 					{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SetVSyncFlag : NI",(long)MipsCpu.PC );
#endif	

		MipsCpu.halt = 1; 
}    





void SYSCALL_NULL( ) { MipsCpu.PC = MipsCpu.PC + 4; }

/*
Execute an Handler at the adress specified by HandlerAdress
NOTA: __inline means compiler do not optimize the funstion
*/
__inline void SYCALL_ExecUserDef() {
	char temp_string[256];
	_u64 SavedRA, SavedPC;
	_u128 		GPR_temp[34];	//BackUp of GeneralRegisters
	_u32	HandlerAdress;
	_u8		SystemCallNumber;
	
 	SystemCallNumber = MipsCpu.GPR[CPU_REG_V1].lo & 0xFF;
 	if ((signed char)SystemCallNumber < 0) SystemCallNumber = (- SystemCallNumber ) & 0xFF ;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SYCALL_UserDef : n %d  at 0x%8.8lx",(long)MipsCpu.PC, 
		(_u8)SystemCallNumber, Ps2Bios.SysCallUserDef[(_u8)SystemCallNumber]);
#endif	


	if ( (_u32)Ps2Bios.SysCallUserDef[(_u8)SystemCallNumber] != (_u32)0xFFFFFFFF ) {



	/* seems ok... at least for the moment */
	MipsCpu.GPR[CPU_REG_V0].lo = INTC_ExecSubFunc( 	(_u32)Ps2Bios.SysCallUserDef[(_u8)SystemCallNumber] ,
									INTC_SUBF_SAVE_REG  , 
									0 , 0, 0);				


#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SYCALL_UserDef END : n %d  at 0x%8.8lx",(long)MipsCpu.PC, 
		SystemCallNumber, Ps2Bios.SysCallUserDef[SystemCallNumber]);
#endif	



	}

	MipsCpu.PC = MipsCpu.PC + 4;
}



/*
SetSyscall
a0: numero
a1: add

*/
void SYSCALL_RFU116( ) 							{ 
	char temp_string[256];
	_u8		SystemCallNumber;
	_u32	SystemCallAdress;

	static _u8 CanSetSyscall = TRUE;

 	

  	SystemCallNumber = MipsCpu.GPR[CPU_REG_A0].lo & 0xFF;
	SystemCallAdress = (_u32)MipsCpu.GPR[CPU_REG_A1].lo ;

 	if ((signed char)SystemCallNumber < 0) SystemCallNumber = (- SystemCallNumber ) & 0xFF ;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU116_SetSyscall : n %d,Addr : 0x%8.8lx ",(long)MipsCpu.PC,
		(_u8)SystemCallNumber, SystemCallAdress );
#endif	


	if ( (_u8)SystemCallNumber == 128 ) 	CanSetSyscall = FALSE;

	
	if( CanSetSyscall == TRUE ) {


	switch( (_u32)SystemCallAdress ) {
	
		case 0:
			syscall_fct_ns[(_u8)SystemCallNumber] = SYSCALL_NULL;
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]RFU116_SetSyscall : ->>  NULL",(long)MipsCpu.PC );
#endif	
		break;


		case 1:
			//syscall_fct_ns[SystemCallNumber] = syscall_fct[SystemCallNumber];
			syscall_fct_ns[(_u8)SystemCallNumber] = SYSCALL_NULL;
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]RFU116_SetSyscall : ->>  NULL",(long)MipsCpu.PC );
#endif	
		break;
		
	
		case 0x80076000:
		case 0x80075000:
			syscall_fct_ns[(_u8)SystemCallNumber] = (void * )SYSCALL_NULL;
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]RFU116_SetSyscall : ->>  NULL",(long)MipsCpu.PC );
#endif	
		break;

		default:
			syscall_fct_ns[(_u8)SystemCallNumber] = (void * )SYCALL_ExecUserDef;
			Ps2Bios.SysCallUserDef[(_u8)SystemCallNumber] = (_u32)SystemCallAdress;
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]RFU116_SetSyscall : n%d->>  SET",(long)MipsCpu.PC ,(_u8)SystemCallNumber);
#endif	
		break;		


	}


	}
	if (( (_u8)SystemCallNumber == 1 ) && (CanSetSyscall == FALSE)) 	CanSetSyscall = TRUE;

    MipsCpu.PC = MipsCpu.PC + 4;
}    


void SYSCALL_print( ) 							{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]print : NI",(long)MipsCpu.PC );
#endif	
	


	MipsCpu.halt = 1; 
}    






void SYSCALL_ExecOSD( ) 						{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ExecOSD",(long)MipsCpu.PC );
#endif	
	
	
/* TEST */


	MipsCpu.GPR[CPU_REG_V0].lo = 0x1;

	MipsCpu.PC = MipsCpu.PC + 4;

}     




void SYSCALL_PSMode( ) 							{ 
	char temp_string[256];
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]PSMode : NI",(long)MipsCpu.PC );
#endif	
	

/*
	sprintf(temp_string,"|PSMODE: A0%8.8lx , A1: %8.8lx\n",(_u32)MipsCpu.GPR[CPU_REG_A0].lo,(_u32)MipsCpu.GPR[CPU_REG_A1].lo);
	PrintToEmuOutput (  temp_string );
*/
	
	//MipsCpu.GPR[CPU_REG_V0].lo = 0x1;

	MipsCpu.halt = 1; 

	//MipsCpu.PC = MipsCpu.PC + 4; 
}     



void SYSCALL_MachineType( ) 					{ 



#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]MachineType : NI",(long)MipsCpu.PC );
#endif	


	MipsCpu.GPR[CPU_REG_V0].lo = 0xFFFFFFFF;

//	MipsCpu.halt = 1; 

	MipsCpu.PC = MipsCpu.PC + 4; 
}     



void SYSCALL_GetMemorySize( ) 					{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]GetMemorySize",(long)MipsCpu.PC );
#endif	




	MipsCpu.GPR[CPU_REG_V0].lo = (_u64)0x02000000;

	MipsCpu.PC = MipsCpu.PC + 4; 
}     









void SYSCALL_iCpuConfig( ) 						{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]iCpuConfig : NI",(long)MipsCpu.PC );
#endif	



	MipsCpu.halt = 1; 
}     


void SYSCALL_iFlushCache( ) 					{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]iFlushCache : NI",(long)MipsCpu.PC );
#endif	


		MipsCpu.halt = 1; 
}     


void SYSCALL_iGetCop0( ) 						{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]iGetCop0 : NI",(long)MipsCpu.PC );
#endif	

	MipsCpu.halt = 1; 
}     




void SYSCALL_RFU091( ) 							{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU091 : NI",(long)MipsCpu.PC );
#endif	
	


//	MipsCpu.halt = 1; 


	MipsCpu.PC = MipsCpu.PC + 4; 
}   













void SYSCALL_UKN_1() { 
	
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SYSCALL_UKN_1 : NI",(long)MipsCpu.PC );
#endif	
	
	MipsCpu.halt = 1;
	//MipsCpu.PC = MipsCpu.PC + 4; 
}    

void SYSCALL_RFU101_105() {
	
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU101_105 : NI",(long)MipsCpu.PC );
#endif	
	
	MipsCpu.halt = 1;


}

void SYSCALL_RFU073( ) 							{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]RFU073 : NI",(long)MipsCpu.PC );
#endif	


	MipsCpu.halt = 1; 
}   



void SYSCALL_SetOsdConfigParam( ) 				{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SetOsdConfigParam : NI",(long)MipsCpu.PC );
#endif	



//MipsCpu.halt = 1; 


INCREMENT_PC();
}   



void SYSCALL_GetOsdConfigParam( ) 				{ 

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]GetOsdConfigParam : NI",(long)MipsCpu.PC );
#endif	


		

	//MipsCpu.halt = 1; 

	INCREMENT_PC();
}   


void SYSCALL_SetOsdConfigParam2( ) {

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SetOsdConfigParam2 : NI",(long)MipsCpu.PC );
#endif	

	
	
	MipsCpu.halt = 1;



}



