/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include <strings.h>
#include <float.h>
#include <math.h>


#include "nSX2.h"

#include "r5900_main.h"
 
#include "bios_syscalls.h"
#include "bios_syscalls_table.h"
 
#include "PS2Edefs.h"

#include "Sif.h"

/* I must declare a function to avoid this kind of include */
#include "DmaC.h"

/* must create function to avoid this kind of declaration */
#include "IntrC.h"


#include "GS.h"



/**********************************************************************************************************
									SEMAPHORE
***********************************************************************************************************/

/*
Create a Semaphore

Param:
A0: address of sema param

Return:
semaphore ID (Normal termination)

*/
void SYSCALL_CreateSema( void ) { 
	char temp_string[256];			//Debug Text	
    _u8 boucle;						//loop var
    
    SemaParam  	*mysm;				//pointer to Semaphore Structure
	SemaParam  	*mysm_alloc;				//pointer to Semaphore Structure
    _s8			FreeSemaRoom;		//ID of a free Semaphore Room  in the BIOS
   
    _u64    TerminationValue;		//Return Value
    
    
   	/* Read the Semaphore Structure */
   	mysm = (  SemaParam * )	&MipsMemory [  MipsCpu.GPR[CPU_REG_A0].lo & 0x01FFFFFF ];
   	
	/* Show Debug Info */

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]CreateSema",(long)MipsCpu.PC );
	
	SYSCALL_LOG("|Address : %llx \n",MipsCpu.GPR[CPU_REG_A0].lo );
	SYSCALL_LOG("|Sema current: %x\n",    	mysm->currentCount);           
	SYSCALL_LOG("|Sema MaxCoun: %x\n",    	mysm->maxCount);		
	SYSCALL_LOG("|Sema Initial: %x\n",    	mysm->initCount);		
	SYSCALL_LOG("|Nb thrds wtg: %x\n",    	mysm->numWaitThreads);	
	SYSCALL_LOG("|Sema attribu: %x\n",    	mysm->attr);			        
	SYSCALL_LOG("|Option: %x\n",    		mysm->option);			        
#endif	
   
	/* Find the first Free Semaphore Room */
	FreeSemaRoom = -1;

	// Get the first Sema room Free
	for ( boucle = 0 ; boucle < BIOS_MAX_SEMAPHORE ; boucle++){
		if ( Ps2Bios.SemaState[ boucle ] == BIOS_SEMA_ROOMFREE ) {
   			FreeSemaRoom = boucle;
    		break;
		}
	}

	// Allocation of Sema Memory
	mysm_alloc = malloc( sizeof(SemaParam) );

	// Copy the Sema Param
	memcpy(mysm_alloc,mysm,sizeof(SemaParam));

#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]CREATE_SEMA : n %d \n",Ps2Bios.PS2_Sched.RunningIDThread,FreeSemaRoom);
#endif

	/* Check If a there is a free room left */
    if ( FreeSemaRoom < BIOS_MAX_SEMAPHORE ) {
    	
    	/* Mark this room as OCCUPIED */
		Ps2Bios.SemaState[ boucle ] = BIOS_SEMA_ROOMBUSY;

		// Set the Sema Param pointer
		Ps2Bios.Sema[ FreeSemaRoom ].sema = mysm_alloc;

		// Initialise the Semaphore values
		Ps2Bios.Sema[ FreeSemaRoom ].sema->currentCount = 	Ps2Bios.Sema[ FreeSemaRoom ].sema->initCount;

		if ( Ps2Bios.Sema[ FreeSemaRoom ].sema->currentCount > Ps2Bios.Sema[ FreeSemaRoom ].sema->maxCount )
			Ps2Bios.Sema[ FreeSemaRoom ].sema->currentCount = Ps2Bios.Sema[ FreeSemaRoom ].sema->maxCount;

		Ps2Bios.Sema[ FreeSemaRoom ].sema->numWaitThreads = 0;

		// Initialise the Semaphore WaitQueue
		InitQueue( Ps2Bios.Sema[ FreeSemaRoom ].WaitQueue);

#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]CREATE_SEMA MaxCount : %d \n",Ps2Bios.PS2_Sched.RunningIDThread,Ps2Bios.Sema[ FreeSemaRoom ].sema->maxCount);
	THREAD_LOG("(%d)[BIOS]CREATE_SEMA InitCount: %d \n",Ps2Bios.PS2_Sched.RunningIDThread,Ps2Bios.Sema[ FreeSemaRoom ].sema->initCount);
	THREAD_LOG("(%d)[BIOS]CREATE_SEMA Cur.Count: %d \n",Ps2Bios.PS2_Sched.RunningIDThread,Ps2Bios.Sema[ FreeSemaRoom ].sema->currentCount);
#endif

    } else {

#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]CREATE_SEMA : Failed \n",Ps2Bios.PS2_Sched.RunningIDThread);
#endif
	}
 	
 	/* Return Value */
    MipsCpu.GPR[CPU_REG_V0].lo = (_u64)FreeSemaRoom;

	INCREMENT_PC();	
} 






/*
Delete Sema

Description:Deletes the specified semaphore. 
An error is returned to threads that have been registered in the
semaphore's wait queue.

Param:
A0 : Semaphore ID

Return:
semaphore ID (Normal termination)
*/
void SYSCALL_DeleteSema( void ){ 
	char temp_string[256];		//Debug Text
	_u8	SemaID;					//Semaphore ID

	_s16 ThreadID;
	LLThread 		* myBiosTh;
	_s16 ReturnValue=-1;

	_s16	boucle;


	_u8		NeedUpdateSceduler = FALSE;


	// Get the Sema ID
 	SemaID = MipsCpu.GPR[CPU_REG_A0].lo ;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]DeleteSema n %d",(long)MipsCpu.PC ,SemaID);
#endif

	if ( Ps2Bios.SemaState[ SemaID ] == BIOS_SEMA_ROOMFREE ) {

		ReturnValue = (_s32)-1;

	} else {

		if ( Ps2Bios.Sema[ SemaID ].sema->numWaitThreads == 0 ) {

#ifdef THREAD_LOG
			THREAD_LOG("(%d)[BIOS]DELETE_SEMA : n %d OK\n",Ps2Bios.PS2_Sched.RunningIDThread,SemaID);
#endif

			// free the BIOS Semaphore room
 			Ps2Bios.SemaState[ SemaID ] = BIOS_SEMA_ROOMFREE;
		
			// free allocated memory
			free( Ps2Bios.Sema[ SemaID ].sema );
			Ps2Bios.Sema[ SemaID ].sema = NULL;
		 

    		ReturnValue = (_u32)SemaID;
 
		} else {
			//numWaitThreads > 0

#ifdef THREAD_LOG
			THREAD_LOG("(%d)[BIOS]DELETE_SEMA Thread present in WaitQueue\n",Ps2Bios.PS2_Sched.RunningIDThread);
#endif

			for ( boucle = 0 ; boucle < Ps2Bios.Sema[ SemaID ].sema->numWaitThreads ; boucle ++ ) {

				// Get the first Thread of the SemaWaitQueue
				ThreadID = Ps2Bios.Sema[ SemaID ].WaitQueue[0];

#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]DELETE_SEMA Error Thread n %d is WakeUp\n",Ps2Bios.PS2_Sched.RunningIDThread,ThreadID);
#endif

				//Get the Thread Param
   				myBiosTh = ( LLThread *) LLThread_Find( Ps2Bios.PS2_Sched.All_Threads, ThreadID );

				//if I wake up the running Thread , I set to RUN
				//else others is set to READY



				// Reset the WaitType and the Id
				myBiosTh->thread->waitType = WAIT_TYPE_NO;
				myBiosTh->thread->waitId = -1;

				THREADS_MoveToReady ( myBiosTh->ID );

				if ( ThreadID == Ps2Bios.PS2_Sched.RunningIDThread ) 
				myBiosTh->thread->status = THS_RUN;
				else 	myBiosTh->thread->status = THS_READY;

			
				//Remove the first Thread of the SemaWaitQueue
				RemoveFirstQueue( Ps2Bios.Sema[ SemaID ].WaitQueue ) ;

		
			}// end of loop

			// free the BIOS Semaphore room
 			Ps2Bios.SemaState[ SemaID ] = BIOS_SEMA_ROOMFREE;
		
			// free allocated memory
			free( Ps2Bios.Sema[ SemaID ].sema );
			Ps2Bios.Sema[ SemaID ].sema = NULL;

    		ReturnValue = (_u32)SemaID;

			// I need to update the Scheduler
			NeedUpdateSceduler = TRUE;
		
		}// end of numWaitThreads > 0

	}

	MipsCpu.GPR[CPU_REG_V0].lo = (_u32)ReturnValue;	

	INCREMENT_PC();	

	if ( NeedUpdateSceduler == TRUE ) UpdatePS2Scheduler();
}  




/*
Signal Semaphore
Description:Release semaphore resource

When the value of sid is 0 and there is no free space in the semaphore queue, 
the thread at the start of the semaphore's wait queue will be released and 
placed in READY state. 
In all other cases, the value of the semaphore is incremented.
For normal termination, the semaphore ID is returned. 
If an invalid sid is specified, a THREAD_ERROR is
returned.

Param:
A0 : Semaphore ID

Return:
semaphore ID (Normal termination)
*/
void SYSCALL_SignalSema( void ) { 
	char temp_string[256];			//Debug Text
	_u32	SemaID;					//Semaphore ID
	_u32	ReturnValue;
	
	LLThread 		* myBiosTh;
	_s16	ThreadID,ThreadID2;

	_u8		NeedUpdateSceduler = FALSE;

 		
 	/* Get Semaphore ID */
	SemaID = MipsCpu.GPR[CPU_REG_A0].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]SignalSema n %d ",(long)MipsCpu.PC ,SemaID);
#endif

	if ( (_s32)Ps2Bios.Sema[ SemaID ].sema->currentCount < 0 ) {
	

	} else {
		
		if ( (_s32)Ps2Bios.Sema[ SemaID ].sema->numWaitThreads > 0 )  {

			// Get the First Thread of the SemaWaitQueue
			ThreadID = Ps2Bios.Sema[ SemaID ].WaitQueue[0];


#ifdef THREAD_LOG
			THREAD_LOG("(%d)[BIOS]SIGNAL_SEMA : Thread  n %d READY \n",Ps2Bios.PS2_Sched.RunningIDThread,ThreadID);
#endif
		
			if ( ThreadID != -1 ) {

				// Get the Thread Param
				myBiosTh = ( LLThread *) LLThread_Find( Ps2Bios.PS2_Sched.All_Threads, ThreadID );

				//Update the Status of the Thread
				
				myBiosTh->thread->waitType = WAIT_TYPE_NO;
				myBiosTh->thread->waitId = -1;

				//Remove the first Thread of the SemaWaitQueue
				RemoveFirstQueue( Ps2Bios.Sema[ SemaID ].WaitQueue ) ;

				//decremente the num of thread waiting on the semaphore
				Ps2Bios.Sema[ SemaID ].sema->numWaitThreads = Ps2Bios.Sema[ SemaID ].sema->numWaitThreads - 1;

				THREADS_MoveToReady ( myBiosTh->ID );		
				myBiosTh->thread->status = THS_READY;

				//I need to update the Scheduler
				NeedUpdateSceduler = TRUE;
			} 

	 	} else {

#ifdef THREAD_LOG
			THREAD_LOG("(%d)[BIOS]SIGNAL_SEMA Inc Sema n %d ( C: %d - T: %d) \n",
					Ps2Bios.PS2_Sched.RunningIDThread,SemaID,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->currentCount,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->numWaitThreads);
#endif

	
			/* the value of the semaphore is incremented */
			Ps2Bios.Sema[ SemaID ].sema->currentCount = Ps2Bios.Sema[ SemaID ].sema->currentCount + 1;
		}

	}


	/*Return Value */
    MipsCpu.GPR[CPU_REG_V0].lo = SemaID;
 
    INCREMENT_PC();	

	if ( NeedUpdateSceduler == TRUE ) UpdatePS2Scheduler();

}   






/*
Wait Semaphore

Description:Get semaphore resource

If the value of the semaphore specified by sid is greater than 1, 
the semaphore is decremented and the function immediately terminates with a normal return. 
If the value of the semaphore is zero, the thread is put in WAIT state and 
placed in the semaphore's wait queue until another thread increments the semaphore by
issuing a SignalSema.

Param:
A0 : Semaphore ID

Return:
semaphore ID (Normal termination)
*/
void SYSCALL_WaitSema( void ) { 
	char 			temp_string[256];				//Debug Text
	_u32			SemaID;							//Semaphore ID
	LLThread 		* myBiosTh;
	_u8				NeedUpdateSceduler = FALSE;
	_u8				AlreadyInQueue;
	_u32			ReturnValue;

	/* Get Semaphore ID */
	SemaID = MipsCpu.GPR[CPU_REG_A0].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]WaitSema n %d",(long)MipsCpu.PC ,SemaID);
#endif

	
	/* if value of the semaphore is greater than 1 */
	if ( (_s32)Ps2Bios.Sema[ SemaID ].sema->currentCount > 0 )  {
 
#ifdef THREAD_LOG
			THREAD_LOG("(%d)[BIOS]WAIT_SEMA Dec Sema n %d ( C: %d - T: %d) \n",
					Ps2Bios.PS2_Sched.RunningIDThread,SemaID,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->currentCount,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->numWaitThreads);
#endif

	 	/* the semaphore is decremented */
	 	Ps2Bios.Sema[ SemaID ].sema->currentCount = Ps2Bios.Sema[ SemaID ].sema->currentCount - 1;
	 
	} else  {

		if ( (_s32)Ps2Bios.Sema[ SemaID ].sema->currentCount < 0 ) { 
		
#ifdef THREAD_LOG
			THREAD_LOG("(%d)[BIOS]WAIT_SEMA ( < 0 ) Sema n %d ( C: %d - T: %d) \n",
					Ps2Bios.PS2_Sched.RunningIDThread,SemaID,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->currentCount,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->numWaitThreads);
#endif
			
 
		} else {

			// currentCount == 0
	 		// the thread is put in WAIT state 
			
			// Get Current running  Thread Param 
			myBiosTh = ( LLThread *) LLThread_Find( Ps2Bios.PS2_Sched.All_Threads, Ps2Bios.PS2_Sched.RunningIDThread );

			// Put the Thread in Wait with WaitType Sema & set the SemaId

			myBiosTh->thread->status = THS_WAIT;
			myBiosTh->thread->waitType = WAIT_TYPE_SEMA;
			myBiosTh->thread->waitId = SemaID;

			
			// I need to update the Scheduler 
			NeedUpdateSceduler = TRUE;
			
			AlreadyInQueue = IsPresentInQueue( Ps2Bios.Sema[ SemaID ].WaitQueue , Ps2Bios.PS2_Sched.RunningIDThread);
			// if the Thread isn't in the Sema WaitQueue
			
			if ( AlreadyInQueue == FALSE ) {

#ifdef THREAD_LOG
				THREAD_LOG("(%d)[BIOS]WAIT_SEMA Thread -> SemaWaitQueue  \n",Ps2Bios.PS2_Sched.RunningIDThread,Ps2Bios.PS2_Sched.RunningIDThread);
#endif
			
				// Increment the number of thread waiting on the Semaphore 
				Ps2Bios.Sema[ SemaID ].sema->numWaitThreads = Ps2Bios.Sema[ SemaID ].sema->numWaitThreads + 1;

				// Add the Current Thread to the Sema Wait Queue 
				AddFirstQueue( Ps2Bios.Sema[ SemaID ].WaitQueue, Ps2Bios.PS2_Sched.RunningIDThread );

		
			} else {
#ifdef THREAD_LOG
				THREAD_LOG("(%d)[BIOS]WAIT_SEMA Thread -> not put in SemaWaitQueue  \n",Ps2Bios.PS2_Sched.RunningIDThread,Ps2Bios.PS2_Sched.RunningIDThread);
#endif
			}
			
		}// end of PUT in Wait State
		
	 } 


 	/* return the Sema ID */
 	MipsCpu.GPR[CPU_REG_V0].lo = (_u32)SemaID;
  	
    INCREMENT_PC();	

	//Test if I need to Update the Scheduler 
	if ( NeedUpdateSceduler == TRUE ) UpdatePS2Scheduler();

}   




/*
The PollSema system call is the same as WaitSema except that the thread does not enter WAIT state.
Unlike WaitSema, an error will be returned if the value of the semaphore is 0.
*/
void SYSCALL_PollSema( ) 						{ 
	char temp_string[256];		//Debug Text
	_u32	SemaID;				//Semaphore ID
	_u32	ReturnValue;

	/* Get Semaphore ID */
	SemaID = MipsCpu.GPR[CPU_REG_A0].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]PollSema n %d",(long)MipsCpu.PC ,SemaID);
#endif


	
	// if value of the semaphore is greater than 1 
	if ( (_s32)Ps2Bios.Sema[ SemaID ].sema->currentCount > 0 ) {


#ifdef THREAD_LOG
			THREAD_LOG("(%d)[BIOS]POLL_SEMA Dec Sema n %d ( C: %d - T: %d) \n",
					Ps2Bios.PS2_Sched.RunningIDThread,SemaID,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->currentCount,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->numWaitThreads);
#endif

		// the semaphore is decremented 
	 	Ps2Bios.Sema[ SemaID ].sema->currentCount = Ps2Bios.Sema[ SemaID ].sema->currentCount - 1;

		// and the return Value is the SemaID 
	 	ReturnValue = (_u32)SemaID;

	} else {

#ifdef THREAD_LOG
			THREAD_LOG("(%d)[BIOS]POLL_SEMA ( <0 ) Sema n %d ( C: %d - T: %d) \n",
					Ps2Bios.PS2_Sched.RunningIDThread,SemaID,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->currentCount,
					(_s32)Ps2Bios.Sema[ SemaID ].sema->numWaitThreads);
#endif

		/* return value is ERROR (-1)*/
		ReturnValue = (_s32)-1;
		
	}

 	/* return value */
    MipsCpu.GPR[CPU_REG_V0].lo = (_u32)ReturnValue;
 
 	
    INCREMENT_PC();	
}   





void SYSCALL_ReferSemaStatus( ) 				{ 

	char temp_string[256];

	_s16	SemaID;
	SemaParam * ParamInfo;

	_u32 InfoParamAdd;	

	/* Get the Sema ID & the Return Add */
	SemaID = (_u32)MipsCpu.GPR[CPU_REG_A0].lo;
	InfoParamAdd = (_u32)MipsCpu.GPR[CPU_REG_A1].lo;

#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]ReferSemaStatus n %d",(long)MipsCpu.PC ,SemaID);
#endif
#ifdef THREAD_LOG
	THREAD_LOG("(%d)[BIOS]REFER_SEMA : %d \n",Ps2Bios.PS2_Sched.RunningIDThread,SemaID );
#endif


	/* Check the Return Addr value */
	if ( InfoParamAdd != 0 ) {

#ifdef THREAD_LOG
		THREAD_LOG("(%d)[BIOS]REFER_SEMA with addr: %8.8x \n",Ps2Bios.PS2_Sched.RunningIDThread,InfoParamAdd );
#endif
		
		/* Get the Addr */
		ParamInfo = (  SemaParam * )	&MipsMemory [  InfoParamAdd & 0x01FFFFFF ];

		/* Copy the specified Sema Param to the return Addr */
		memcpy ( ParamInfo, Ps2Bios.Sema[ SemaID ].sema, sizeof(SemaParam) );
			
	}

	MipsCpu.PC = MipsCpu.PC + 4; 

}   


void SYSCALL_iReferSemaStatus( ) 				{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]iReferSemaStatus : NI",(long)MipsCpu.PC );
#endif

	MipsCpu.halt = 1; 
}     


void SYSCALL_iPollSema( ) 						{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]iPollSema",(long)MipsCpu.PC );
#endif

	SYSCALL_PollSema();
}     



void SYSCALL_iSignalSema( ) 					{ 
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]iSignalSema",(long)MipsCpu.PC );
#endif

	SYSCALL_SignalSema();
}     

