/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/


/*
*/
/********************************************************************************************
							GENERAL EMULATOR DEFINITION
*********************************************************************************************/

#ifndef _NSX2_H
#define _NSX2_H


#include <stdio.h>
#include <stdlib.h>
#include "nSX2_type.h"

#include "bios.h"
#include "DmaC.h"
#include "IntrC.h"
#include "timers.h"
#include "MemMan.h"
#include "VIF.h"
#include "GIF.h"
#include "VU.h"


#define nSX2_VERSION_MAJ	0
#define nSX2_VERSION_MIN	8
#define nSX2_VERSION_BUILD	0
 


#define	nSX2ConfigFile	"./nsx2.ini"

#define nSX2DemoDir		"./demos/"

/*
Use the GS pluggin
*/
#define PLUGINtypedefs
#define PLUGINfuncs




/*
Set to 1, for use scph10000.bin
*/
#define	USE_PS2BIOS		0



#define	EMU_LOG	AddLogLine
//#undef	EMU_LOG




/********************************************************************************************
							TYPE DEFINITION
*********************************************************************************************/
#ifndef TRUE

#define TRUE 	1
#define FALSE 	0

#endif


/********************************************************************************************
							EMULATOR DEFINES
*********************************************************************************************/

#define	LOAD_CDVD		0
#define	LOAD_ELF_FILE	1
#define	LOAD_NOTHING	2	
#define	LOAD_BIOS		4	



#define CPU_REG_R0	0
#define CPU_REG_AT	1
#define CPU_REG_V0	2
#define CPU_REG_V1	3
#define CPU_REG_A0	4
#define CPU_REG_A1	5
#define CPU_REG_A2	6
#define CPU_REG_A3	7
#define CPU_REG_T0	8
#define CPU_REG_T1	9
#define CPU_REG_T2	10
#define CPU_REG_T3	11
#define CPU_REG_T4	12
#define CPU_REG_T5	13
#define CPU_REG_T6	14
#define CPU_REG_T7	15
#define CPU_REG_S0	16
#define CPU_REG_S1	17
#define CPU_REG_S2	18
#define CPU_REG_S3	19
#define CPU_REG_S4	20
#define CPU_REG_S5	21
#define CPU_REG_S6	22
#define CPU_REG_S7	23
#define CPU_REG_T8	24
#define CPU_REG_T9	25
#define CPU_REG_K0	26
#define CPU_REG_K1	27
#define CPU_REG_GP	28
#define CPU_REG_SP	29
#define CPU_REG_S8	30
#define CPU_REG_RA	31
#define CPU_REG_HI	32
#define CPU_REG_LO	33

#define SIGN_BIT_MASK       0x80000000
#define SIGN_BIT_MASK_64    0x8000000000000000

/*
COP 1 Control Register
*/
#define CP1_FIR		0	//Floating-Point Implementation Register (FIR, CP1 Control Register 0)
#define CP1_FCCR	25	//Floating-Point Condition Codes Register (FCCR, CP1 Control Register 25)
#define CP1_FEXR	26	//Floating-Point Exceptions Register (FEXR, CP1 Control Register 26)
#define CP1_FENR	28	//Floating-Point Enables Register (FENR, CP1 Control Register 28)
#define CP1_FCSR	31	//Floating-Point Control and Status Register (FCSR, CP1 Control Register 31)

#define CP1_FCSR_C_MASK     0x00800000
#define CP1_FCSR_C_NOT_MASK 0xFF7FFFFF


/********************************************************************************************
							STRUCTURES DEFINITIONS
*********************************************************************************************/




/*
Main CPU R5900
*/
typedef	struct _Mips_CPU {
	_u64		PC;						//Program Counter	
	_u32		opcode;					//current opcode
	_u8			halt;					//if 1 stop the emu

	

	_u8			InterpRunning;
	
	_u128 		GPR[34];				//General Registers
	_u128 		SA;

	_u64		CPR0[32][4];			//Cop0	Control Register

	_u32		FPR1[32];				//COP1	FloatingPoint Register for the CPU
    float		*FPR1_p_float[32];		//COP1	Pointer Float for conversion

	_u32		FPR1_ACC;
	float		*FPR1_ACC_p_float;		
	
	_u64		CP1[32];				//COP1 Control Register
	
	_u64		Nb_instructions;		//Number Of Instruction Exececuted

	_u32 		EntryPointAdd;			//Start the emu at


	_u64		Nb_Cycles;
	
	_u32		vuOpcode;


	
} Mips_CPU;










/*
Pad
*/

/*
 * Pad states
 */
#define PAD_STATE_DISCONN     	0x00
#define PAD_STATE_FINDPAD       0x01
#define PAD_STATE_FINDCTP1      0x02
#define PAD_STATE_EXECCMD       0x05
#define PAD_STATE_STABLE        0x06
#define PAD_STATE_ERROR      	0x07



/*
 * Pad request states
 */
#define PAD_RSTAT_COMPLETE      0x00
#define PAD_RSTAT_FAILED        0x01
#define PAD_RSTAT_BUSY          0x02





/*
typedef struct _Pad_State
{
    int open;
    unsigned int port;
    unsigned int slot;
    struct pad_data *padData;
    unsigned char *padBuf;
}Pad_State;
*/

typedef struct _Pad_Data 
{
    unsigned int frame;			//0-3
    unsigned char state;		//4
    unsigned char reqState;		//5
    unsigned char ok;			//6
    unsigned char unkn7;		//7
    unsigned char data[32];		//8-39
    unsigned int length;		//40-43
    unsigned int unkn44;		//44-47
    unsigned int unkn48;		//48		: Pad Type
    unsigned int unkn52;
    unsigned int unkn54;
    unsigned int unkn60;
}Pad_Data ;

typedef struct _padButtonStatus
{
    unsigned char ok;			//0
    unsigned char mode;			//1
    unsigned char btns[2];		//2-3
    // joysticks
    unsigned char rjoy_h;		//4
    unsigned char rjoy_v;		//5
    unsigned char ljoy_h;		//6
    unsigned char ljoy_v;		//7
    // pressure mode
    unsigned char left_p;		//8
    unsigned char rigth_p;		//9
    unsigned char up_p;			//10
    unsigned char down_p;		//11
    unsigned char triangle_p;	//12
    unsigned char circle_p;		//13
    unsigned char square_p;		//14
    unsigned char cross_p;		//15
    unsigned char unkn16[16];	//16-31
}padButtonStatus;

/* rom0:xpadman */
#define XPAD_BIND_RPC_ID1 0x80000100
#define XPAD_BIND_RPC_ID2 0x80000101


typedef struct _XPad_Data 
{
    unsigned char data[32]; // 0, length = 32 bytes
    unsigned int unkn32;    // not used??
    unsigned int unkn36;    // not used??
    unsigned int unkn40;    // byte 40  not used??
    unsigned int unkn44;    // not used?? 44
    unsigned char actData[32]; // actuator (6x4?) 48
    unsigned short modeTable[4];  // padInfo   80
    unsigned int frame;     // byte 88, u32 22
    unsigned int unkn92;    // not used ??
    unsigned int length;    // 96
    unsigned char modeOk;   // padInfo  100 Dunno about the name though...
    unsigned char modeCurId; // padInfo    101
    unsigned char unkn102;  // not used??
    unsigned char unknown;  // unknown
    unsigned char nrOfModes;   // padInfo   104
    unsigned char modeCurOffs; // padInfo   105
    unsigned char nrOfActuators;     // act  106
    unsigned char unkn107[5];  // not used??
    unsigned char state;    // byte 112
    unsigned char reqState; // byte 113
    unsigned char ok;       // padInfo  114
    unsigned char unkn115[13];  // not used??
}XPad_Data;

/* end of pad.c */

/* end of padlib 1.2 */


/*
Internal Emulator BIOS
*/

#define		BIOS_MAX_SEMAPHORE		32
#define		BIOS_SEMA_ROOMFREE		0
#define		BIOS_SEMA_ROOMBUSY		1




/*
This Structure define 
BIOS PS2 
and 
Other Hardware : DMA Controller, Timer, etc

*/
typedef	struct _Bios_Hw {


	/* SEMAPHORE STUFF */
	Bios_Sema	Sema		[BIOS_MAX_SEMAPHORE];
	_u8			SemaState	[BIOS_MAX_SEMAPHORE];
	
	/* THREADS STUFF*/
	Bios_Scheduler	PS2_Sched;
	PS2_Thread 		InitThread;

	/* SYSCALL USER DEF */
	_u32	SysCallUserDef[256];

	/* Pads */
	Pad_Data 	* PadData[2][2];
	XPad_Data 	* XPadData[2][2];
	padButtonStatus 	* PadState[2];
	
	/* VERY BAD WAY*/
	_u64	NeedUpdateScreen;
	_u8		AccesToGS;

	_u8 	NeedUpdatePad;

	_u64	NbOfInstructionsPerSec;

} Bios_Hw;





/********************************************************************************************
							EXTERN GLOBALS VARIABLES
*********************************************************************************************/



extern FILE * ElfFile;		//Current Elf File
extern long ElfFileSize;	//Size of the Elf File
extern FILE * BiosFile;		//Current BIOS File


extern Mips_CPU MipsCpu;	//Main CPU
extern Bios_Hw  Ps2Bios;	//Internal BIOS


extern _u8	*MipsMemory;			//Main Memory 32 Mbits
extern _u8	*SpecialAreaMemory;		//Special Memory ( PS2 Control Register  )
extern _u8	*GSAreaMemory;			//GS control register
extern _u8	*BiosMemory;			//Bios Memory
extern _u8  *SPRMemory;

extern void (*syscall_fct_ns[256])(void);	//non-static

#define		FILE_NAME_LENGHT		1024




extern _u32  	(* PS2EgetLibTypeGS)();
extern _u32  	(* PS2EgetLibVersionGS)();
extern char*	(* PS2EgetLibNameGS)();



extern _u32  	(* PS2EgetLibTypePAD)();
extern _u32  	(* PS2EgetLibVersionPAD)();
extern char*	(* PS2EgetLibNamePAD)();




extern	_u8		EnableLogFile;

/********************************************************************************************
							MACROS DEFINITIONS
*********************************************************************************************/
  





/*******************************************************************************************/
/*
void muGUI_nSX2Proc ( void ) ;
void muGUI_nSX2InitProc ( void ) ;
void muGUI_nSX2EndingProc ( void ) ;
*/





/********************************************************************************************
							CONFIG FILE DECLARATIONS
*********************************************************************************************/
int 	nSX2LoadConfig 	(void);

extern char cCurDir[MAX_PATH];

/********************************************************************************************
							MAIN CPU R5900 DECLARATIONS
*********************************************************************************************/
void	MipsUpdateCpu 		(void);
void	MipsCpuPrefetchOpcode(void);

extern _u8		debug_op;
extern _u8		debug_regs;

//extern _u8		debug_print_syscall;
extern _u8		debug_systemcall;

extern _u8		debug_gs;
extern _u8		debug_dma;
extern _u8		debug_gif;
extern _u8		debug_intc;
extern _u8		debug_timer;

extern _u8		debug_vif;
extern _u8		debug_thread ;

extern _u8      debug_file ;

extern _u8      GS_VSync_Wait ; 
/********************************************************************************************
							BIOS DECLARATIONS
*********************************************************************************************/


void BiosUpdate( void );





/********************************************************************************************
							SIF RPC DECLARATIONS
*********************************************************************************************/

#include "SIF_RPC.h"

#define	SIF_PLUGGIN_MAX_BIND	32

#define		BIND_RPC_NPM_USER	0
#define		BIND_CDVD_IRX		1
#define		BIND_SJPCM_IRX		2	
#define		BIND_AMIGAMOD_IRX	3
	
#define		BIND_0x80000001		4
#define		BIND_0x80000003		5
#define		BIND_0x80000400		6
#define		BIND_0x80000592		7
#define		BIND_0x80000593		8
#define		BIND_0x80000595		9
#define		BIND_0x80000596		10
#define		BIND_0x80000597		11
#define		BIND_0x8000059a		12
#define		BIND_0x80000601		13
		
#define		BIND_0x80000006		14
#define		BIND_0x8000010f		15
#define		BIND_0x8000011f		16
#define		BIND_0x80000100		17
#define		BIND_0x80000101		18




#define PAD_BIND_RPC_ID1 	0x8000010f
#define PAD_BIND_RPC_ID2 	0x8000011f


#define AMIGAMOD_IRX  		0xbadca11
#define	SJPCM_IRX			0xB0110C5
#define RPC_NPM_USER 		0x014d704e
#define CDVD_IRX			0xB001337

void RPC_8000010f_PadMan1	( _u32 fRPCcommand, _u64 fBinaryInAddr, _u64 fBinaryOutAddr ) ;




void PadAndKeyUpdate( void ) ;


float ComputeInstructionsPerSec( void  );



#endif
