/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include <float.h>


#include "nSX2.h"
 
#include "r5900_main.h"
 

#include "r5900_debug_regs.h"

#include "r5900_core.h"	//for NI

#include "r5900_cop1.h"
#include "r5900_cop1_table.h"

#include "PS2Edefs.h"



/*
Pofis Add Opcodes :
C_F_S, CFC1, MIN_S, SUBA_S


*/

/*****************************************************************************************************************
									COP1 INSTRUCTION SET
*****************************************************************************************************************/

void    COP1	()		{	cop1_opc[( _OPCODE_ & 0x3E00000)>>21 ](); 	}

void	BC1		()		{ 	cop1_bc1_opc[( _OPCODE_ >>16)& 0x3 ](); 	}

void	COP1_S	()		{ 	cop1_s_opc[ _OPCODE_ & 0x3F ](); 			}

void	COP1_W	()		{ 	cop1_W_L_opc[ _OPCODE_ & 0x3F ](); 			}


void	ABS_S	( void )	{	
          
    *MipsCpu.FPR1_p_float[CURRENT_FD] = (float)fabs( (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) );
                
	INCREMENT_PC();
} 	



void	ADD_S	( void )	{	
          
    *MipsCpu.FPR1_p_float[CURRENT_FD] = (float)((float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) + (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));
                
	INCREMENT_PC();
} 


void 	ADDA_S	( void )	{	
           
    *MipsCpu.FPR1_ACC_p_float = (float)((float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) + (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));
                
	INCREMENT_PC();
} 



void	BC1F	( void )	{	
	_u64	target_branch;
	_u32	condition;

	_u64 off;



	condition = MipsCpu.CP1[CP1_FCSR] & CP1_FCSR_C_MASK;

	if ( condition == 0 ) {
		
		if ( debug_op == 1 ) PrintToOpcodeOutput ( debugTempText );

		CalculTargetBranch;
		
		INCREMENT_PC();
		MipsUpdateCpu();
		
		MipsCpu.PC = target_branch;
		
	}
	else INCREMENT_PC(); 
} 




void	BC1FL	( void )	{	
	_u64	target_branch;
	_u32	condition;
	_u64 off;


	condition = MipsCpu.CP1[CP1_FCSR] & CP1_FCSR_C_MASK;

	if ( condition == 0 ) {
		
		if ( debug_op == 1 ) PrintToOpcodeOutput ( debugTempText );

		CalculTargetBranch;
		
		INCREMENT_PC();
		MipsUpdateCpu();
		
		MipsCpu.PC = target_branch;
		
	}
	else MipsCpu.PC = MipsCpu.PC + 8; 
} 




void	BC1T	( void )	{	
	_u64	target_branch;
	_u32	condition;

	_u64 off;


	condition = MipsCpu.CP1[CP1_FCSR] & CP1_FCSR_C_MASK;

	if ( condition == CP1_FCSR_C_MASK ) {
		
		if ( debug_op == 1 ) PrintToOpcodeOutput ( debugTempText );

		CalculTargetBranch;
		
		INCREMENT_PC();
		MipsUpdateCpu();
		
		MipsCpu.PC = target_branch;
	}
	else INCREMENT_PC();
} 


void	BC1TL	( void )	{	
	_u64	target_branch;
	_u32	condition;

	_u64 off;
	
	condition = MipsCpu.CP1[CP1_FCSR] & CP1_FCSR_C_MASK;
	
	if ( condition == CP1_FCSR_C_MASK ) {
		
		if ( debug_op == 1 ) PrintToOpcodeOutput ( debugTempText );

		CalculTargetBranch;
		
		INCREMENT_PC();
		MipsUpdateCpu();
		
		MipsCpu.PC = target_branch;
	}
	else MipsCpu.PC = MipsCpu.PC + 8; 
}




void	C_EQ_S	( void )	{	

	if ( (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) == (float) (*MipsCpu.FPR1_p_float[CURRENT_FT])  ) 
		MipsCpu.CP1[CP1_FCSR] = MipsCpu.CP1[CP1_FCSR] | CP1_FCSR_C_MASK;
	else  MipsCpu.CP1[CP1_FCSR] = MipsCpu.CP1[CP1_FCSR] & CP1_FCSR_C_NOT_MASK;
 
  	INCREMENT_PC();
}	




void	C_F_S	( void )	{	

	MipsCpu.CP1[CP1_FCSR] = MipsCpu.CP1[CP1_FCSR] & CP1_FCSR_C_NOT_MASK;
  	INCREMENT_PC();
} 	




void	C_OLE_S	( void )	{	

	if ( (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) <= (float)(*MipsCpu.FPR1_p_float[CURRENT_FT])  )
        MipsCpu.CP1[CP1_FCSR] = MipsCpu.CP1[CP1_FCSR] | CP1_FCSR_C_MASK;
	else MipsCpu.CP1[CP1_FCSR] = MipsCpu.CP1[CP1_FCSR] & CP1_FCSR_C_NOT_MASK;

	INCREMENT_PC();
} 


void	C_OLT_S	( void )	{	

	if ( (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) < (float)(*MipsCpu.FPR1_p_float[CURRENT_FT])  )
        MipsCpu.CP1[CP1_FCSR] = MipsCpu.CP1[CP1_FCSR] | CP1_FCSR_C_MASK;
	else MipsCpu.CP1[CP1_FCSR] = MipsCpu.CP1[CP1_FCSR] & CP1_FCSR_C_NOT_MASK;

	INCREMENT_PC();
}	



void	CFC1	( void )	{	

	MipsCpu.GPR[CURRENT_RT].lo = (_s64)((_s32)MipsCpu.CP1[CURRENT_FS]);
	INCREMENT_PC();
} 		



void	CTC1	( void )	{	

	MipsCpu.CP1[CURRENT_FS] = (_u32)(MipsCpu.GPR[CURRENT_FT].lo);
	INCREMENT_PC();
} 	


void	CVTS_W	( void )	{	

	*MipsCpu.FPR1_p_float[CURRENT_FD] = (float)((_s32)MipsCpu.FPR1[CURRENT_FS]);

  	INCREMENT_PC();
}



void	CVT_W_S	( void )	{	
    _u16    f_exp;
    
    f_exp = (_u16)( (MipsCpu.FPR1[CURRENT_FS] >> 23) & 0xFF) ;
     
    if ( (_u16) f_exp <= 0x9D ) {
      MipsCpu.FPR1[CURRENT_FD] =(_s32)((float)*MipsCpu.FPR1_p_float[CURRENT_FS]);
    }
    else {
        if ( (MipsCpu.FPR1[CURRENT_FS] & SIGN_BIT_MASK ) == 0 ) MipsCpu.FPR1[CURRENT_FD] = (_u32)0x7FFFFFFF;
        else MipsCpu.FPR1[CURRENT_FD] = (_u32)0x80000000;
   }

   INCREMENT_PC();
}	


void	DIV_S	( void )	{	


	if ( ( (*MipsCpu.FPR1_p_float[CURRENT_FS]) != 0.0 ) && ( (*MipsCpu.FPR1_p_float[CURRENT_FT]) == 0.0 )  ) {
        
        if ( ( (*MipsCpu.FPR1_p_float[CURRENT_FS]) != 0.0 ) > 0.0 ) *MipsCpu.FPR1_p_float[CURRENT_FD] = FLT_MAX;
        else *MipsCpu.FPR1_p_float[CURRENT_FD] = FLT_MIN;
    }

	if ( ( (*MipsCpu.FPR1_p_float[CURRENT_FS]) == 0.0 ) && ( (*MipsCpu.FPR1_p_float[CURRENT_FT]) == 0.0 )  ) {
        *MipsCpu.FPR1_p_float[CURRENT_FD] = FLT_MAX;
    }

	if ( ( (*MipsCpu.FPR1_p_float[CURRENT_FS]) != 0.0 ) && ( (*MipsCpu.FPR1_p_float[CURRENT_FT]) != 0.0 )  ) {
	
   		*MipsCpu.FPR1_p_float[CURRENT_FD] = (float)((float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) / (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));
   
    } 

/*
	if (  (*MipsCpu.FPR1_p_float[CURRENT_FT]) != 0.0 )  {
	
   		*MipsCpu.FPR1_p_float[CURRENT_FD] = (float)((float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) / (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));
   
    } 
*/

	INCREMENT_PC();
} 	



void    LWC1	( void )	{	
	_u128	temp;
	_u64 off;
	
	off = Extractoffset();
	sign_extendedh(off);

	
	MipsMemoryRead ( &temp , DATA_W ,  MipsCpu.GPR[CURRENT_BASE].lo , off ) ;

	MipsCpu.FPR1[CURRENT_FT] = (_u32)temp.lo ;
	
    INCREMENT_PC();
}



void 	MADD_S	( void )	{	

	(*MipsCpu.FPR1_p_float[CURRENT_FD])	 =   (float)(*MipsCpu.FPR1_ACC_p_float) + ((float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) * (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));

    INCREMENT_PC();
}



void 	MADDA_S	( void )	{	

	(*MipsCpu.FPR1_ACC_p_float)	 =   (float)(*MipsCpu.FPR1_ACC_p_float) + ((float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) * (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));

    INCREMENT_PC();
} 


void 	MAX_S	( void )	{	


	if ( (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) >=  (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]) ) {

		(*MipsCpu.FPR1_p_float[CURRENT_FD]) = (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]);

	} else {

		(*MipsCpu.FPR1_p_float[CURRENT_FD]) = (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]);
	}

    INCREMENT_PC();
}	




void	MFC1	( void )	{	

	MipsCpu.GPR[CURRENT_RT].lo = (_s64)((_s32)MipsCpu.FPR1[CURRENT_FS]);

    INCREMENT_PC();
}


void	MIN_S	( void )	{	

	if ( (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) <=  (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]) ) {

		(*MipsCpu.FPR1_p_float[CURRENT_FD]) = (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]);

	} else {

		(*MipsCpu.FPR1_p_float[CURRENT_FD]) = (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]);
	}

    INCREMENT_PC();

} 



void	MOV_S	( void )	{	

    *MipsCpu.FPR1_p_float[CURRENT_FD] = (float) (*MipsCpu.FPR1_p_float[CURRENT_FS]);

	INCREMENT_PC();
} 




void 	MSUB_S	( void )	{	

	(*MipsCpu.FPR1_p_float[CURRENT_FD])	 =   (float)*MipsCpu.FPR1_ACC_p_float - ((float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) * (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));

    INCREMENT_PC();



}


void 	MSUBA_S	( void )	{	

	(*MipsCpu.FPR1_ACC_p_float) = (float)*MipsCpu.FPR1_ACC_p_float - ((float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) * (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));
	INCREMENT_PC();

} 



void	MTC1	( void )	{	
	
	MipsCpu.FPR1[CURRENT_FS] = (_u32)( MipsCpu.GPR[CURRENT_RT].lo  );						//Copy the val to the reg
	INCREMENT_PC();	
}	



void	MUL_S	( void )	{	

	*MipsCpu.FPR1_p_float[CURRENT_FD] =(float)( (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) * (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));
   
 	INCREMENT_PC();
} 



void	MULA_S	( void )	{	

	(*MipsCpu.FPR1_ACC_p_float) =(float)( (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) * (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));
   
 	INCREMENT_PC();


} 
 

void	NEG_S	( void )	{	


    *MipsCpu.FPR1_p_float[CURRENT_FD] = (float)(-(*MipsCpu.FPR1_p_float[CURRENT_FS]));
    
	INCREMENT_PC();
}


void	RSQRT_S	( void )	{	


	*MipsCpu.FPR1_p_float[CURRENT_FD] = ((float) *MipsCpu.FPR1_p_float[CURRENT_FS]) / ((float)sqrt( *MipsCpu.FPR1_p_float[CURRENT_FT] ));
	INCREMENT_PC();
}



void	SQRT_S	( void )	{	
	
	if ( *MipsCpu.FPR1_p_float[CURRENT_FT] >= 0.0f )
		*MipsCpu.FPR1_p_float[CURRENT_FD] = (float)sqrt( *MipsCpu.FPR1_p_float[CURRENT_FT] );

	else 
		*MipsCpu.FPR1_p_float[CURRENT_FD] = sqrt( -(*MipsCpu.FPR1_p_float[CURRENT_FT]) );
	
	INCREMENT_PC();
}		



void	SUB_S	( void )	{

	*MipsCpu.FPR1_p_float[CURRENT_FD] = (float)((float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) - (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]));

	INCREMENT_PC();
} 		

		


void 	SUBA_S	( void )	{	

	(*MipsCpu.FPR1_ACC_p_float) = (float)(*MipsCpu.FPR1_p_float[CURRENT_FS]) - (float)(*MipsCpu.FPR1_p_float[CURRENT_FT]);

	INCREMENT_PC();

} 



void    SWC1	( void )	{	
	_u128 temp128;
	_u64 off;
	off = Extractoffset();
	sign_extendedh(off);
	
	temp128.lo  = (_u32)MipsCpu.FPR1[CURRENT_FT];

	MipsMemoryWrite ( temp128 , DATA_W ,  MipsCpu.GPR[CURRENT_BASE].lo , off ) ;

    INCREMENT_PC();
}





