/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#include "nSX2_Type.h"

#ifndef _TIMERS_H
#define _TIMERS_H


/* Define or Undefine if you want  TIMERS_LOG */
#define	TIMERS_LOG	TIMERS_Log
//#undef	TIMERS_LOG

#define	TIMERS_LOG_REG	TIMERS_Log
//#undef	TIMERS_LOG_REG

/********************************************************************************/
/*						TIMERS DEFINES											*/
/********************************************************************************/


#define	PS2_TIMER_0			0
#define	PS2_TIMER_1			1
#define	PS2_TIMER_2			2
#define	PS2_TIMER_3			3

#define	NSX2_TIMER_GS		4
#define	NSX2_TIMER_INPUT	5
#define	NSX2_MAIN_TIMER		6
#define	NSX2_THREAD_TIMER	7




/********************************************************************************/
/*						TIMERS REGISTER DEFINES									*/
/********************************************************************************/

#define		T0_COUNT	0x10000000		//Timer 0 Counter value
#define		T0_MODE		0x10000010		//Timer 0 Mode/Status
#define		T0_COMP		0x10000020		//Timer 0 Compare value
#define		T0_HOLD		0x10000030		//Timer 0 Hold Value

#define		T1_COUNT	0x10000800		//Timer 1 Counter value
#define		T1_MODE		0x10000810		//Timer 1 Mode/Status
#define		T1_COMP		0x10000820		//Timer 1 Compare value
#define		T1_HOLD		0x10000830 		//Timer 1 Hold Value

#define		T2_COUNT	0x10001000		//Timer 2 Count
#define		T2_MODE		0x10001010		//Timer 2 Mode/status
#define		T2_COMP		0x10001020		//Timer 2 Compare value

#define		T3_COUNT	0x10001800		//Timer 3 Counter value
#define		T3_MODE		0x10001810		//Timer 3 Mode/status
#define		T3_COMP		0x10001820		//Timer 3 Compare value



/********************************************************************************/
/*						TIMERS REGISTER STRUCTURE								*/
/********************************************************************************/


typedef struct _TimersReg {
	_u32	count;		// 0x00	- Timer Counter value	
	_u32	null0[3];	// 0x04 0x08 0x0C
	_u32	mode;		// 0x10	- Timer	Mode/Status	
	_u32	null1[3];	// 0x14 0x18 0x1C
	_u32	comp;		// 0x20	- Timer Compare value	
	_u32	null2[3];	// 0x24 0x28 0x2C
	_u32	hold;		// 0x30	- Timer	Hold Value	
}TimersReg;


/********************************************************************************/
/*						TIMERS  STRUCTURE										*/
/********************************************************************************/


typedef	struct _STimer {

	TimersReg * Reg;

	_u8	running;

}STimer;



/********************************************************************************/
/*						MACRO DEFINITIONS										*/
/********************************************************************************/


#define		GET_MODE_CLKS_TIMER( A ) 	(   Timer[ (A) ].Reg->mode        & 0x3 )
#define		GET_MODE_GATE_TIMER( A ) 	( ( Timer[ (A) ].Reg->mode >> 2 ) & 0x1 )
#define		GET_MODE_GATS_TIMER( A ) 	( ( Timer[ (A) ].Reg->mode >> 3 ) & 0x1 )
#define		GET_MODE_GATM_TIMER( A ) 	( ( Timer[ (A) ].Reg->mode >> 4 ) & 0x3 )
#define		GET_MODE_ZRET_TIMER( A ) 	( ( Timer[ (A) ].Reg->mode >> 6 ) & 0x1 )
#define		GET_MODE_CUE_TIMER( A ) 	( ( Timer[ (A) ].Reg->mode >> 7 ) & 0x1 )
#define		GET_MODE_CMPE_TIMER( A ) 	( ( Timer[ (A) ].Reg->mode >> 8 ) & 0x1 )
#define		GET_MODE_OVFE_TIMER( A ) 	( ( Timer[ (A) ].Reg->mode >> 9 ) & 0x1 )
#define		GET_MODE_EQUF_TIMER( A ) 	( ( Timer[ (A) ].Reg->mode >> 10 ) & 0x1 )
#define		GET_MODE_OVFF_TIMER( A ) 	( ( Timer[ (A) ].Reg->mode >> 11 ) & 0x1 )


#define		SET_MODE_EQUF_TIMER( A )	Timer[ (A) ].Reg->mode = Timer[ (A) ].Reg->mode | 0x400
#define		SET_MODE_OVFF_TIMER( A )	Timer[ (A) ].Reg->mode = Timer[ (A) ].Reg->mode | 0x800

#define		CLEAR_MODE_EQUF_TIMER( A )	Timer[ (A) ].Reg->mode = Timer[ (A) ].Reg->mode & (~0x400)
#define		CLEAR_MODE_OVFF_TIMER( A )	Timer[ (A) ].Reg->mode = Timer[ (A) ].Reg->mode & (~0x800)


/********************************************************************************/
/*						GLOBAL VARIABLES										*/
/********************************************************************************/


extern STimer Timer[ 8 ];
extern _u64	TIMERS_PreviousCycles;


/********************************************************************************/
/*						FUNCTIONS DECLARATIONS									*/
/********************************************************************************/


/* Main */
void 	TIMERS_Log 		( char * ftext  , ...   );
void	TIMERS_Init		( void );
void	TIMERS_Write 	( _u64 fPhysicalAddress, _u128 fVal, 	_u8 fDataType );
void	TIMERS_Read 	( _u64 fPhysicalAddress, _u128 * fVal, 	_u8 fDataType );
void 	TIMERS_Update	( void );


/* Common */
void TIMERS_TestOverlow 	( _u8 fTimer );
void TIMERS_TestEqual 		( _u8 fTimer );
void TIMERS_TestZeroReturn 	( _u8 fTimer );

/* Cycles */
_u8 Timers_GetCycles		( _u32 fOpcode );

/* Emu */
void TIMERS_Update_GS 		( void );
void TIMERS_Update_Input 	( void );
void TIMERS_Update_Thread 	( void );


void TIMERS_PS2_Update 		( _u8 fTimer );

#endif





/*

humm.. let's go for mathematical...

f = 300 Mhz

( First problem :
I don't remenber if 300Mhz is 300000000 Hz or
if 300 MHz = 300 * 1024 * 1024 = 314572800 Hz
For the following , I take 300000000
)

f = 1 / T ( f:frequency , T:Period )

T = 1 / f
unit is : 	T ( second ) 			f ( Hz )
			T ( milliSecond )		f ( KHz - 10e3 Hz )
			T ( microSecond )		f ( MHz - 10e6 Hz )

T 	= 1 / 300 = 0,00333333.... microSecond
	= 3,333 nanoSecond ( 10e-9 )

if f = 300MHz,		1 cycle = 3,33 nanoSecond		( CLOCK = 300 Mhz)
if f = 150 MHz, 	1 cycle = 6,66 nanoSecond		( BUSCLOCK = 150 Mhz )
if f = 9,375 MHz,   1 cycle = 106,56 nanoSecond 	( BUSCLOCK/16 = 9,375 )
if f = 0,585 Mhz,	1 cycle = 1706,66 nanoSecond	( BUSCLOCK/256 = 0,585 )


the Reference CLOCk is 300MHz
BUSCLOCK  		==>  2 cycle
BUSCLOCK/16		==>  32 cycle
BUSCLOCK/256	==>  512 cycle




the EE runs at 300MHZ!
if I want update the VSync 25 times per seconds,
1 frame = 1 sec / 25 = 0,04 sec
0,04 sec = 40 milliSecond =  40 000 microSecond = 40 000 000 nanoSecond 
40 000 000 / 3.33 = 12 012 012 cycles


for the HSync
if the resolution is 640  *  240 , the HSync occurs 240 times by frame
12 012 012 / 240 = 50050 cycles
the HSyns occurs each 50050 cycles


*/

















